"""Sending and receiving messages between QMI objects and between QMI contexts.

The classes defined in this module are for the most part used internally by QMI.
Application programs should typically not interact with these classes directly.
"""

import sys
import asyncio
import copy
import fnmatch
import functools
import logging
import os
import pickle
import random
import socket
import threading
import time

from typing import Callable, Dict, Generic, List, NamedTuple, Optional, Tuple, TypeVar

import qmi

from qmi.core.thread import QMI_Thread
from qmi.core.exceptions import (
    QMI_Exception, QMI_RuntimeException, QMI_TimeoutException, QMI_InvalidOperationException,
    QMI_MessageDeliveryException, QMI_UsageException,
    QMI_DuplicateNameException, QMI_UnknownNameException
    )

from qmi.core.udp_responder_packets import (
        unpack_qmi_udp_packet,
        QMI_UdpResponderContextInfoRequestPacket,
        QMI_UdpResponderContextInfoResponsePacket,
        QMI_UdpResponderKillRequestPacket,
    )

from qmi.core.util import format_address_and_port, parse_address_and_port


# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


# Full address of a message handler, including its context name.
class QMI_MessageHandlerAddress(NamedTuple):
    """Unique address associated with a specific message handler.

    Attributes:
        context_id: Name of the context that contains the message handler.
        object_id: Unique name for the message handler within the context.
            In many cases, this is simply the RPC object name.
    """

    context_id: str
    object_id: str

    def __str__(self) -> str:
        return self.context_id + "." + self.object_id


class QMI_Message:
    """Base class for all messages sent via QMI.

    Attributes:
        source_address: Address of the object which sent this message.
        destination_address: Address of the object which should receive this message.
    """

    __slots__ = ("source_address", "destination_address")

    def __init__(self,
                 source_address: QMI_MessageHandlerAddress,
                 destination_address: QMI_MessageHandlerAddress
                 ) -> None:
        self.source_address = source_address
        self.destination_address = destination_address


class QMI_RequestMessage(QMI_Message):
    """Base class for request messages in QMI.

    Request messages are used when a reply is expected.
    When an object sends a request message, it expects to receive exactly
    one corresponding reply message.

    The association between request messages and reply messages is tracked
    explicitly via the `request_id` attribute.

    Attributes:
        request_id: Unique ID for this message.
            Automatically initialized to a random 64-bit integer.
    """

    __slots__ = ("request_id",)

    def __init__(self,
                 source_address: QMI_MessageHandlerAddress,
                 destination_address: QMI_MessageHandlerAddress
                 ) -> None:
        super().__init__(source_address, destination_address)
        self.request_id = "{:016x}".format(random.getrandbits(64))


class QMI_ReplyMessage(QMI_Message):
    """Base class for reply messages in QMI.

    A reply message is a response to a request message.
    When an object receives a request message, it is expected to send
    exactly one corresponding reply message.

    The destination of the reply message is taken from the source
    of the corresponding request message. The `request_id` attribute
    of the reply message must be equal to the `request_id` of the
    corresponding request message.

    Attributes:
        request_id: Request ID of the `QMI_RequestMessage` to which this message is a reply.
    """

    __slots__ = ("request_id",)

    def __init__(self,
                 source_address: QMI_MessageHandlerAddress,
                 destination_address: QMI_MessageHandlerAddress,
                 request_id: str
                 ) -> None:
        super().__init__(source_address, destination_address)
        self.request_id = request_id


class QMI_ErrorReplyMessage(QMI_ReplyMessage):
    """Special reply message, sent when a request can not be processed.

    An instance of `QMI_ErrorReplyMessage` is generated when a request message
    can not be delivered, or when the associated connection is lost before
    a reply is received. This message will then be delivered in place of
    the actual reply to the pending request.

    Attributes:
        error_msg: Short string, describing the reason why the request could not be processed.
    """

    __slots__ = ("error_msg",)

    def __init__(self,
                 source_address: QMI_MessageHandlerAddress,
                 destination_address: QMI_MessageHandlerAddress,
                 request_id: str,
                 error_msg: str
                 ) -> None:
        super().__init__(source_address, destination_address, request_id)
        self.error_msg = error_msg


class QMI_InitialHandshakeMessage(QMI_Message):
    """Initial handshake message exchanged for a new TCP connection."""

    __slots__ = ("version", "is_server_handshake")

    def __init__(self, context_name: str, version: str, is_server_handshake: bool) -> None:
        source_address = QMI_MessageHandlerAddress(context_name, "$router")
        # QMI_InitialHandshakeMessage does not use a destination address.
        destination_address = QMI_MessageHandlerAddress("", "")
        super().__init__(source_address, destination_address)
        self.version = version
        self.is_server_handshake = is_server_handshake


class QMI_MessageHandler:
    """Base class for QMI classes which can receive messages.

    A message handler instance has a unique address and is registered
    with the QMI context.
    """

    def __init__(self, address: QMI_MessageHandlerAddress) -> None:
        self.address = address

    def handle_message(self, message: QMI_Message) -> None:
        """Called by the QMI message router when a message for this object is received.

        This method is responsible for handling the received message.
        This method should not raise exceptions other than `QMI_MessageDeliveryException`.

        Subclasses must implement this method.

        Raises:
            QMI_MessageDeliveryException: If the message can not be accepted.
        """
        raise NotImplementedError()

    def shutdown(self) -> None:
        """Called by the QMI message router during unregistering of the handler.

        After this call, no more messages will be received.

        Subclasses can implement this method to free resources they are using.
        The default implementation does nothing.

        Do not call this method explicitly. It will be called automatically
        during unregistering of the message handler.
        """
        pass


_T = TypeVar('_T')


class _EventDrivenThread(QMI_Thread):
    """An EventDrivenThread executes an Asyncio event loop."""

    class Future(Generic[_T]):
        """Representation of the future completion of a procedure running in an `EventDrivenThread`."""

        def __init__(self, thread: '_EventDrivenThread') -> None:
            self._thread = thread
            self._finished = False
            self._value = None  # type: Optional[_T]
            self._exception = None  # type: Optional[Exception]

        def set_value(self, value: _T) -> None:
            with self._thread._cv:
                assert not self._finished
                self._value = value
                self._finished = True
                self._thread._cv.notify_all()

        def set_exception(self, exc: Exception) -> None:
            with self._thread._cv:
                assert not self._finished
                self._exception = exc
                self._finished = True
                self._thread._cv.notify_all()

        def wait(self, timeout: Optional[float] = None) -> _T:
            with self._thread._cv:
                ok = self._thread._cv.wait_for(lambda: (self._thread._thread_finished or self._finished),
                                               timeout=timeout)
            if not ok:
                raise QMI_TimeoutException()
            if not self._finished:
                raise QMI_InvalidOperationException("Thread stopped before operation completed")
            if self._exception is not None:
                raise self._exception
            return self._value  # type: ignore

    def __init__(self) -> None:
        super().__init__()
        self._cv = threading.Condition()
        self._thread_initialized = False
        self._thread_finished = False

        # self.event_loop holds this thread's Asyncio event loop.
        # This variable is only valid while the thread runs
        # (i.e. _thread_initialized and not _thread_finished) and may only
        # be accessed by code running inside the thread.
        self.event_loop = None  # type: Optional[asyncio.AbstractEventLoop]

    def run(self) -> None:

        _logger.debug("EventDrivenThread starting")

        # Create Asyncio event loop for this thread.
        # It has to be of type SelectorEventLoop.
        # The default event loop type on Windows is ProactorEventLoop,
        # but that does not support the socket notifier mechanism we use.
        self.event_loop = asyncio.SelectorEventLoop()
        asyncio.set_event_loop(self.event_loop)

        # Mark thread as initialized.
        with self._cv:
            self._thread_initialized = True
            self._cv.notify_all()

        try:
            # Run event loop until event_loop.stop() is called.
            self.event_loop.run_forever()

        finally:
            # Mark thread as finished.
            with self._cv:
                self._thread_finished = True
                self._cv.notify_all()

            # Clean up event loop.
            self.event_loop.close()
            self.event_loop = None

        _logger.debug("EventDrivenThread finished")

    def _request_shutdown(self) -> None:
        """Initiate termination of the thread.

        This is called once during `shutdown()` of the thread.
        """

        assert self._shutdown_requested

        # Check that thread was started (i.e. self.start() was called).
        # If the thread is not currently running and has not completed
        # its initialization, it means start() was never called and
        # this is an invalid use case (calling shutdown() before start()).
        assert self.is_alive() or self._thread_initialized

        with self._cv:

            # Wait until thread is initialized.
            while not self._thread_initialized:
                self._cv.wait()

            # Tell the event loop to stop, unless the thread is already finished.
            if not self._thread_finished:
                assert self.event_loop is not None
                self.event_loop.call_soon_threadsafe(self.event_loop.stop)

    def run_in_thread_arg(self, func: Callable[[_T], None], arg: _T) -> None:
        """Run the specified function inside the thread.

        This method returns immediately. The specified function will run
        inside the thread after some short but indeterminate delay.

        The thread must be started before calling this method.
        If the thread is stopped before the specified function gets a chance
        to run, nothing happens.

        Parameters:
            func: Function to run in the thread.
            arg: A single argument to be passed to the function.
        """

        # Check that thread was started (i.e. self.start() was called).
        assert self.is_alive() or self._thread_initialized

        with self._cv:

            # Wait until thread is initialized.
            while not self._thread_initialized:
                self._cv.wait()

            # Tell the event loop to run the specified function.
            if not self._thread_finished:
                assert self.event_loop is not None
                self.event_loop.call_soon_threadsafe(func, arg)

    @staticmethod
    def _call_no_arg_helper(func: Callable[[], None]) -> None:
        func()

    def run_in_thread(self, func: Callable[[], None]) -> None:
        """Run the specified function inside the thread.

        This method returns immediately. The specified function will run
        inside the thread after some short but indeterminate delay.

        The thread must be started before calling this method.
        If the thread is stopped before the specified function gets a chance
        to run, nothing happens.

        Parameters:
            func: Function to run in the thread.
        """
        self.run_in_thread_arg(self._call_no_arg_helper, func)

    def make_future(self) -> '_EventDrivenThread.Future':
        return _EventDrivenThread.Future(self)

    @staticmethod
    def _call_wait_helper(func: Callable[[], _T], future: Future[_T]) -> None:
        try:
            ret = func()
            future.set_value(ret)
        except Exception as exc:
            future.set_exception(exc)

    def run_in_thread_wait(self, func: Callable[[], _T]) -> _T:
        """Run the specified function inside the thread and wait until it completes.

        This method waits while the specified function runs in the thread,
        then returns the value returned by the specified function.

        The thread must already be started before calling this method.
        If the thread is stopped before the specified function gets a chance
        to run, QMI_InvalidOperationException is raised.

        Parameters:
            func: Function to run in the thread.

        Returns:
            Value returned by the specified function.

        Raises:
            QMI_InvalidOperationException: If the thread stops before the function runs.
            Exception: If the specified function raises the same exception.
        """

        # Check that thread was started (i.e. self.start() was called).
        assert self.is_alive() or self._thread_initialized

        with self._cv:

            # Wait until thread is initialized.
            while not self._thread_initialized:
                self._cv.wait()

            # Check that the thread is still running.
            if self._thread_finished:
                raise QMI_InvalidOperationException("EventDrivenThread already stopped before run_in_thread_wait()")

            # Tell the event loop to run the specified function.
            assert self.event_loop is not None
            future = self.make_future()  # type: _EventDrivenThread.Future[_T]
            self.event_loop.call_soon_threadsafe(self._call_wait_helper, func, future)

            # Wait until the function completes and return its result.
            return future.wait()


class _SocketWrapper:
    """Base class for socket wrappers.

    A SocketWrapper encapsulates a socket object and handles read events on the socket.

    This is an abstract base class. Subclasses implement actual behavior for specific use cases.
    """

    def close(self) -> None:
        """Close the wrapped socket object and release resources.

        Subclasses must override this method.
        """
        raise NotImplementedError()


class _UdpResponder(_SocketWrapper):
    """Handles requests received via a dedicated UDP socket.

    The UdpResponder runs in an Asyncio event loop to respond to asynchronous
    events on the UDP socket.

    Instances of `UdpResponder` are managed by the `SocketManager`.
    """

    def __init__(self,
                 event_loop: asyncio.AbstractEventLoop,
                 message_router: 'MessageRouter',
                 sock: socket.socket
                 ) -> None:
        """Initialize a UdpResponder instance.

        Parameters:
            event_loop: Asyncio event loop to which the socket should attach itself.
            message_router: `MessageRouter` instance that manages this UDP responder.
            sock: UDP server socket, already bound and listening.
        """

        super().__init__()
        self._event_loop = event_loop
        self._message_router = message_router
        self._sock = sock

        # POSIX sockets should not block on recv_from(), when select() returns True. However,
        # in Linux, this is not always true. We make the socket non-blocking and
        # and explicitly handle recv() failure to make up for this.
        self._sock.setblocking(False)

        # Register callback to be invoked when the socket is ready for reading.
        self._event_loop.add_reader(self._sock.fileno(), self._handle_read)

        _logger.debug("UDP responder ready on %s", format_address_and_port(self._sock.getsockname()))

    def close(self) -> None:
        """Detach from the event loop and close the wrapped socket."""
        _logger.debug("Closing UDP responder")
        self._event_loop.remove_reader(self._sock.fileno())
        self._sock.close()

    def _handle_read(self) -> None:
        """Called through the event loop when the socket may be ready for reading."""

        try:
            (request_data, incoming_address) = self._sock.recvfrom(4096)
        except BlockingIOError:
            # No UDP message available. Continue.
            _logger.error("UdpResponder got handle_read but no packet available")
            return

        # Try to interpret it as a QMI packet.
        # If that fails, discard the packet.
        try:
            request_packet = unpack_qmi_udp_packet(request_data)
        except QMI_Exception:
            _logger.exception("Discarded bad UDP packet.")
            return

        # Handle the different kinds of incoming packets.
        if isinstance(request_packet, QMI_UdpResponderContextInfoRequestPacket):
            self._handle_context_info_request_packet(incoming_address, request_packet)
        elif isinstance(request_packet, QMI_UdpResponderKillRequestPacket):
            self._handle_kill_request_packet()
        else:
            _logger.warning("Received UDP packet of type %s, discarded.", type(request_packet).__name__)

    def _handle_context_info_request_packet(self,
                                            incoming_address: Tuple[str, int],
                                            request_packet: QMI_UdpResponderContextInfoRequestPacket
                                            ) -> None:
        """Called when a `QMI_UdpResponderContextInfoRequestPacket` is received."""

        # First, we check if we want to respond, by checking the request's
        # "workgroup_name_filter" and "context_name_filter" to the corresponding values of the Context.

        workgroup_name_filter = request_packet.workgroup_name_filter.decode()
        workgroup_name_match = fnmatch.fnmatchcase(self._message_router.workgroup_name, workgroup_name_filter)

        if not workgroup_name_match:
            # Don't respond.
            return

        context_name_filter = request_packet.context_name_filter.decode()
        context_name_match = fnmatch.fnmatchcase(self._message_router.context_name, context_name_filter)

        if not context_name_match:
            # Don't respond.
            return

        # Send response.

        response_msg_id = random.randint(1, 2 ** 64 - 1)  # Positive 64-bit integer.

        response_packet = QMI_UdpResponderContextInfoResponsePacket.create(
            response_msg_id,  # Unique message ID.
            time.time(),  # POSIX timestamp.
            request_packet.pkt_id,
            request_packet.pkt_timestamp,
            os.getpid(),
            self._message_router.context_name.encode(),
            self._message_router.workgroup_name.encode(),
            self._message_router.tcp_server_port
        )
        outgoing_address = incoming_address
        self._sock.sendto(bytes(response_packet), outgoing_address)

    def _handle_kill_request_packet(self) -> None:
        """Called when a `QMI_UdpResponderKillRequestPacket` is received."""
        print("Answering external hard-kill request, exiting with exitcode 1.")
        os._exit(1)


class _PeerTcpConnection(_SocketWrapper):
    """Encapsulates a TCP connection to a peer context.

    The TCP connection can either be an outgoing (client-side) connection to a
    remote server, or an incoming (server-side) connection from a remote client.

    The `PeerTcpConnection` is initially a passive object. It will later be attached
    to an Asyncio event loop to enable background processing of received data.

    Instances of `PeerTcpConnection` are managed by the `SocketManager`.
    """

    # Maximum size of serialized message is 10 MB.
    MAX_MESSAGE_SIZE = 10000000

    def __init__(self,
                 message_router: 'MessageRouter',
                 sock: socket.socket,
                 peer_context_alias: str,
                 is_incoming: bool
                 ) -> None:
        """Initialize a PeerTcpConnection instance.

        Parameters:
            message_router: `MessageRouter` instance that manages this TCP connection.
            sock: TCP socket, already connected to a remote TCP socket.
            peer_context_alias: Local name for the peer context. This is either the actual name
                of the remote context (for outgoing connections) or a local unique identifier
                (for incoming connections).
            is_incoming: True if this is an incoming connection from a remote client;
                False if this is an outgoing connection to a remote server.
        """
        self._message_router = message_router
        self._sock = sock
        self.peer_context_alias = peer_context_alias
        self._is_incoming = is_incoming
        self._event_loop = None  # type: Optional[asyncio.AbstractEventLoop]
        self._socket_manager = None  # type: Optional[_SocketManager]
        self._recv_buf = bytearray()

        # Actual peer context name initially unknown (until handshake).
        self.peer_context_name = None  # type: Optional[str]

        # Actual peer context version initially unknown (until handshake).
        self.peer_context_version = None  # type: Optional[str]

        # Table of pending outgoing request messages, by request ID.
        self._pending_requests: Dict[str, Tuple[QMI_MessageHandlerAddress, QMI_MessageHandlerAddress]] = {}

        # NOTE: Do not make TCP socket non-blocking (in contrast to UDP socket)
        # because sending to the TCP socket should be a blocking operation.

        self.local_address = sock.getsockname()
        self.peer_address = None  # type: Optional[Tuple[str, int]]
        self._peer_address_str = "[unknown]"
        try:
            self.peer_address = sock.getpeername()
            self._peer_address_str = format_address_and_port(self.peer_address)
        except OSError:
            # Maybe connection already broke - then we don't know the peer address.
            pass

    def attach_to_socket_manager(self,
                                 event_loop: asyncio.AbstractEventLoop,
                                 socket_manager: '_SocketManager'
                                 ) -> None:
        """Attach this PeerTcpConnection to the socket manager and start
        event-driven processing of received data.
        """
        assert self._event_loop is None
        assert self._socket_manager is None

        self._event_loop = event_loop
        self._socket_manager = socket_manager

        # Register callback to be invoked when the socket is ready for reading.
        self._event_loop.add_reader(self._sock.fileno(), self._handle_read)

    def close(self) -> None:
        """Detach from the event loop and close the wrapped socket."""
        _logger.debug("Closing TCP connection to %s (%r)",
                      self._peer_address_str,
                      self.peer_context_name)
        if self._event_loop is not None:
            self._event_loop.remove_reader(self._sock.fileno())
        self._sock.close()
        self._clear_pending_requests()

    def _clear_pending_requests(self) -> None:
        """Send failure replies for any pending requests on this socket.

        This function is called when the connection is closed.
        There may be tasks in the local context that are still waiting
        for replies to requests that were sent out via this connection.
        Because the socket is now closed, these replies will never be received.

        For each such pending request, we generate an error reply to notify
        the waiting task that the request will not be answered.
        """
        for (request_id, (source_address, destination_address)) in self._pending_requests.items():
            reply = QMI_ErrorReplyMessage(
                source_address=destination_address,
                destination_address=source_address,
                request_id=request_id,
                error_msg="Connection to {} closed while waiting for reply".format(self.peer_context_name))
            try:
                self._message_router.deliver_message(reply)
            except QMI_MessageDeliveryException:
                _logger.debug("Failed to deliver error reply to %r while closing socket", source_address)
        self._pending_requests.clear()

    def _handle_read(self) -> None:
        """Called through the event loop when the socket is ready for reading."""

        assert self._socket_manager is not None

        # If this function raises an exception, it will kill the Asyncio event
        # loop (and thus the whole SocketManager). DO NOT WANT!
        # Catch any such exceptions here and close the connection.
        try:
            self._receive_data()
        except Exception as exc:
            _logger.info("Error on connection to %s (%s) - closing",
                         self._peer_address_str,
                         self.peer_context_name,
                         exc_info=True)
            # Close connection.
            self._socket_manager.remove_peer_connection(self)
            self.close()

    def _receive_data(self) -> None:
        """Called when the socket is ready for reading."""

        assert self._socket_manager is not None

        incoming_data = self._sock.recv(4096)
        if len(incoming_data) == 0:
            # Connection closed by other side.
            _logger.debug("Connection to %s (%s) closed by peer", self._peer_address_str, self.peer_context_name)

            # Close our side as well and remove from the socket manager.
            self._socket_manager.remove_peer_connection(self)
            self.close()
            return

        # Add received data to buffer.
        self._recv_buf.extend(incoming_data)

        # Consume any complete messages from the receive buffer.
        while True:

            if len(self._recv_buf) < 1:
                break

            if self._recv_buf[0] != ord(b'P'):
                raise QMI_RuntimeException("Protocol violation (got {!r} while expecting 'P')"
                                           .format(self._recv_buf[0:1]))

            if len(self._recv_buf) < 9:
                break

            pickled_message_size = int.from_bytes(self._recv_buf[1:9], byteorder='little')

            if pickled_message_size > self.MAX_MESSAGE_SIZE:
                # Protocol violation.
                raise QMI_RuntimeException('Protocol packet too big ({})'.format(pickled_message_size))

            if len(self._recv_buf) < 9 + pickled_message_size:
                break

            packed_message = self._recv_buf[9:9 + pickled_message_size]
            self._recv_buf = self._recv_buf[9 + pickled_message_size:]

            self._process_message(packed_message)

    def _process_message(self, packed_message: bytearray) -> None:
        """Called when a message has been received from the socket."""

        # This may fail.
        message = pickle.loads(packed_message)
        if not isinstance(message, QMI_Message):
            raise ValueError("Expected QMI_Message")

        if self.peer_context_name is None:

            # The first message from the peer must be a handshake message.
            if not isinstance(message, QMI_InitialHandshakeMessage):
                raise QMI_RuntimeException("Expecting handshake message but got {}"
                                           .format(type(message).__name__))

            # Process handshake message.
            self.peer_context_name = message.source_address.context_id
            self.peer_context_version = message.version
            if message.is_server_handshake and self._is_incoming:
                raise QMI_RuntimeException("Received server handshake from connecting client")
            if (not message.is_server_handshake) and (not self._is_incoming):
                raise QMI_RuntimeException("Received client handshake while connecting as client")
            _logger.debug("Received handshake from %r", message.source_address.context_id)

        else:
            if isinstance(message, QMI_InitialHandshakeMessage):
                raise QMI_RuntimeException("Unexpected handshake message from peer")

            # Check destination context name.
            if message.destination_address.context_id != self._message_router.context_name:
                raise QMI_MessageDeliveryException(
                    "Unexpected destination context {} in message from {}"
                    .format(message.destination_address.context_id, self.peer_context_name))

            # Check source context name and rewrite to local alias.
            if message.source_address.context_id != self.peer_context_name:
                raise QMI_MessageDeliveryException("Unexpected source context {} in message from {}"
                                                   .format(message.source_address.context_id, self.peer_context_name))
            message.source_address = QMI_MessageHandlerAddress(
                self.peer_context_alias,
                message.source_address.object_id
            )

            # In case of a reply message, remove the corresponding request from the table.
            if isinstance(message, QMI_ReplyMessage):
                if message.request_id in self._pending_requests:
                    self._pending_requests.pop(message.request_id)
                else:
                    _logger.warning("Received reply message for unknown request_id %r", message.request_id)

            # Push message to router for local delivery.
            try:
                self._message_router.deliver_message(message)
            except QMI_MessageDeliveryException as exc:
                # Message delivery failed.
                # If this is a request message, send back and error reply.
                _logger.warning("%s", str(exc))
                if isinstance(message, QMI_RequestMessage):
                    self.send_error_reply(message, str(exc))
            except Exception:
                # This happens if a "handle_message()" method raises an unexpected exception.
                _logger.exception("Unexpected exception while delivering message to %r", message.destination_address)

    def send_message(self, message: QMI_Message) -> None:
        """Send a message to the peer context via this TCP connection.

        This function serializes the message and sends it through TCP.
        An exception is raised if serialization or transmission fails.

        When this function sends a request message, the request is added
        to the pending request table, to be cleared when a reply is received.
        """

        # Replace local name of destination context by actual context name.
        # Skip this step only for QMI_InitialHandshakeMessage, which does not have a destination.
        message = copy.copy(message)
        if not isinstance(message, QMI_InitialHandshakeMessage):
            assert message.destination_address.context_id == self.peer_context_alias
            assert self.peer_context_name is not None
            message.destination_address = QMI_MessageHandlerAddress(
                self.peer_context_name,
                message.destination_address.object_id
            )

        # Serialize the message.
        pickled_message = pickle.dumps(message)

        pickled_message_size = len(pickled_message)
        if pickled_message_size > self.MAX_MESSAGE_SIZE:
            raise ValueError("Message exceeds maximum size")

        # Prepend the header.
        pickled_message = b'P' + pickled_message_size.to_bytes(8, byteorder='little') + pickled_message

        # Send the message via TCP.
        self._sock.sendall(pickled_message)

        # In case of a request message, add the message to the pending request table.
        if isinstance(message, QMI_RequestMessage):
            if message.request_id in self._pending_requests:
                _logger.warning("Duplicate request_id %r in message to %r",
                                message.request_id,
                                message.destination_address)
            else:
                self._pending_requests[message.request_id] = (message.source_address, message.destination_address)

    def send_handshake(self) -> None:
        """Send an initial handshake message to the remote side."""
        message = QMI_InitialHandshakeMessage(
            self._message_router.context_name,
            qmi.__version__,
            self._is_incoming
        )
        self.send_message(message)

    def send_error_reply(self, message: QMI_RequestMessage, error_msg: str) -> None:
        """Send an error reply message back to the peer context via this TCP connection."""
        reply = QMI_ErrorReplyMessage(
            source_address=message.destination_address,
            destination_address=message.source_address,
            request_id=message.request_id,
            error_msg=error_msg)
        try:
            self.send_message(reply)
        except (ValueError, OSError):
            _logger.warning("Error while sending error reply to %s", message.destination_address.context_id)

    def receive_handshake(self, timeout: Optional[float]) -> None:
        """Wait (blocking) until the socket receives a handshake message.

        This function is used when creating a client-side connection,
        to verify the handshake in the main thread.
        """

        assert self._event_loop is None
        assert self.peer_context_name is None

        if timeout is None:
            endtime = None
        else:
            endtime = time.monotonic() + timeout

        while True:

            # Check message header.
            if (len(self._recv_buf) > 0) and (self._recv_buf[0] != ord(b'P')):
                raise QMI_RuntimeException("Protocol violation (got {!r} while expecting 'P')"
                                           .format(self._recv_buf[0:1]))

            # Decode message length.
            if len(self._recv_buf) < 9:
                need_len = 9
            else:
                pickled_message_size = int.from_bytes(self._recv_buf[1:9], byteorder='little')
                if pickled_message_size > self.MAX_MESSAGE_SIZE:
                    raise QMI_RuntimeException('Protocol packet too big ({})'.format(pickled_message_size))
                need_len = 9 + pickled_message_size

            # Stop when message complete.
            if len(self._recv_buf) >= need_len:
                break

            # Receive additional bytes.
            if endtime is not None:
                tmo = max(0, endtime - time.monotonic())
                self._sock.settimeout(tmo)
            incoming_data = self._sock.recv(need_len - len(self._recv_buf))
            if len(incoming_data) == 0:
                # Connection closed by other side.
                raise QMI_RuntimeException("Connection to {} closed by peer before handshake"
                                           .format(self._peer_address_str))

            self._recv_buf.extend(incoming_data)

        if endtime is not None:
            # Put socket back in unconditional blocking mode.
            self._sock.settimeout(None)

        packed_message = self._recv_buf[9:9 + pickled_message_size]
        self._recv_buf = self._recv_buf[9 + pickled_message_size:]

        # This must be the handshake (since we had not receive a handshake yet).
        self._process_message(packed_message)

        # We must now know the remote peer name (from the handshake).
        assert self.peer_context_name is not None


class _TcpServer(_SocketWrapper):
    """Encapsulates a TCP server socket.

    A TcpServer runs in an Asyncio event loop to respond to asynchronous
    connection events on the TCP server socket.

    Instances of `TcpServer` are managed by the `SocketManager`.
    """

    def __init__(self,
                 event_loop: asyncio.AbstractEventLoop,
                 socket_manager: '_SocketManager',
                 sock: socket.socket
                 ) -> None:
        """Initialize a TcpServer instance.

        Parameters:
            event_loop: Asyncio event loop to which the socket should attach itself.
            message_router: `MessageRouter` instance that manages this UDP responder.
            sock: TCP server socket, already bound and listening.
        """

        self._event_loop = event_loop
        self._socket_manager = socket_manager
        self._sock = sock

        # Make the socket non-blocking.
        self._sock.setblocking(False)

        # Register callback to be invoked when the socket is ready for reading.
        self._event_loop.add_reader(self._sock.fileno(), self._handle_read)

        _logger.debug("TCP server ready on %s", format_address_and_port(self._sock.getsockname()))

    def close(self) -> None:
        """Detach from the event loop and close the wrapped socket."""
        _logger.debug("TCP server on %s closing", format_address_and_port(self._sock.getsockname()))
        self._event_loop.remove_reader(self._sock.fileno())
        self._sock.close()

    def _handle_read(self) -> None:
        """Called through the event loop when a new TCP connection is established."""

        # Accept new connection.
        try:
            (incoming_connection_socket, incoming_connection_address) = self._sock.accept()
        except OSError as exc:
            # Accept can fail in case of early TCP protocol error.
            _logger.exception("Accepting new TCP connection failed")
            return

        _logger.debug("Incoming TCP connection from %s", format_address_and_port(incoming_connection_address))

        # Disable TCP small packet delay (Nagle algorithm).
        try:
            incoming_connection_socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        except OSError:
            _logger.exception("Setting TCP_NODELAY failed on incoming TCP connection")
            # Continue normal flow; we can in principle work correctly without TCP_NODELAY.

        # Add new connection to socket manager.
        self._socket_manager.add_incoming_connection(incoming_connection_socket)


class _SocketManager:
    """The `SocketManager` manages the UDP and TCP sockets of a context.

    The `SocketManager` manages one UDP responder socket, an optional TCP
    server socket, and possibly multiple TCP peer connection sockets.

    The `SocketManager` runs in a separate thread in an event-driven fashion.
    Unless stated otherwise, the public methods of the `SocketManager` may only
    be called in the socket manager thread.
    """

    def __init__(self, event_loop: asyncio.AbstractEventLoop, message_router: 'MessageRouter') -> None:
        _logger.debug("SocketManager initializing")
        self._event_loop = event_loop
        self._message_router = message_router
        self._lock = threading.Lock()
        self._socket_wrappers = []  # type: List[_SocketWrapper]
        self._peer_context_map = {}  # type: Dict[str, _PeerTcpConnection]
        self._peer_name_counter = 0

    def close_all(self) -> None:
        """Close all sockets managed by the SocketManager."""
        _logger.debug("SocketManager closing all sockets")
        with self._lock:
            self._peer_context_map.clear()
        for sock in self._socket_wrappers:
            sock.close()
        self._socket_wrappers.clear()

    def add_tcp_server(self, sock: socket.socket) -> None:
        """Add a TCP server socket."""
        tcp_server = _TcpServer(self._event_loop, self, sock)
        self._socket_wrappers.append(tcp_server)

    def add_udp_responder(self, sock: socket.socket) -> None:
        """Add an UDP responder socket."""
        udp_responder = _UdpResponder(self._event_loop, self._message_router, sock)
        self._socket_wrappers.append(udp_responder)

    def add_incoming_connection(self, sock: socket.socket) -> None:
        """Add an incoming TCP connection from a remote client context.

        The TCP connection is already accepted.
        A handshake with the client still needs to be performed.
        """

        # Assign unique "anonymous" name for the remote side of this connection.
        self._peer_name_counter += 1
        peer_context_alias = "$client_{}".format(self._peer_name_counter)

        # Wrap socket in PeerTcpConnection object.
        conn = _PeerTcpConnection(self._message_router,
                                  sock,
                                  peer_context_alias,
                                  is_incoming=True)

        # Send handshake and attach to event loop.
        try:
            conn.attach_to_socket_manager(self._event_loop, self)
            conn.send_handshake()
        except Exception as exc:
            _logger.exception("Error on new incoming connection")
            conn.close()
            return

        # Add connection to index of managed sockets.
        self._socket_wrappers.append(conn)
        with self._lock:
            self._peer_context_map[conn.peer_context_alias] = conn

    def add_outgoing_connection(self, conn: _PeerTcpConnection) -> None:
        """Add an outgoing TCP connection to a remote TCP server.

        The TCP connection is already established and the handshake has been completed.
        There must not be an existing connection to a remote peer context with the same name.
        """

        # Check against duplicate peer context name (the MessageRouter already prevents it, but we double-check).
        assert conn.peer_context_alias not in self._peer_context_map

        # Check that the handshake is complete.
        assert conn.peer_context_name is not None

        # Attach connection to the event loop.
        conn.attach_to_socket_manager(self._event_loop, self)

        # Add connection to index of managed sockets.
        self._socket_wrappers.append(conn)
        with self._lock:
            self._peer_context_map[conn.peer_context_alias] = conn

        # Notify MessageRouter that a new peer context was connected.
        self._message_router.notify_peer_context_added(conn.peer_context_alias)

    def remove_peer_connection(self, conn: _PeerTcpConnection) -> None:
        """Remove the specified peer connection from the socket manager.

        This method is called by _PeerTcpConnectionSocket when it detects
        that its connection to the peer context is no longer valid.
        """
        with self._lock:
            peer_context_alias = conn.peer_context_alias
            assert self._peer_context_map[peer_context_alias] is conn
            self._peer_context_map.pop(peer_context_alias)
        self._socket_wrappers.remove(conn)

        # Notify MessageRouter that peer context became disconnected.
        self._message_router.notify_peer_context_removed(peer_context_alias)

    def disconnect_from_peer(self, peer_context_name: str) -> None:
        """Disconnect from the specified peer context."""
        conn = self._peer_context_map.get(peer_context_name)
        if conn is None:
            raise QMI_UnknownNameException("Unknown peer context {}".format(peer_context_name))
        self.remove_peer_connection(conn)
        conn.close()

    def send_message(self, message: QMI_Message) -> None:
        """Send the specified message to the right peer context.

        If the destination context is not one of the currently connected
        peer contexts, the message will be dropped. If this happens with
        a request message, a corresponding error reply message will be
        generated.
        """

        # NOTE: This function must not raise exceptions, because any exceptions
        #   raised here would end up in the top level Asyncio event loop.

        assert message.source_address.context_id == self._message_router.context_name

        error_msg: Optional[str] = None

        # Determine which connection to use to send this message.
        destination_context_name = message.destination_address.context_id
        conn = self._peer_context_map.get(destination_context_name)

        if conn is not None:
            try:
                conn.send_message(message)
            except (ValueError, OSError) as exc:
                # TODO qmi#379 - It sometimes happens that a background service
                #     logs 1000s of BrokenPipeError exceptions within 1 second.
                #     To be investigated why this happens.
                _logger.exception(
                    "Error while sending message from %s.%s to %s.%s type %s",
                    message.source_address.context_id,
                    message.source_address.object_id,
                    message.destination_address.context_id,
                    message.destination_address.object_id,
                    type(message))
                error_msg = "{}: {}".format(type(exc).__name__, str(exc))
        else:
            _logger.warning("Unknown message destination context %r", destination_context_name)
            error_msg = "Unknown message destination context {}".format(destination_context_name)

        # Generate an error reply if we fail to route a request message.
        if (error_msg is not None) and isinstance(message, QMI_RequestMessage):
            reply = QMI_ErrorReplyMessage(source_address=message.destination_address,
                                          destination_address=message.source_address,
                                          request_id=message.request_id,
                                          error_msg=error_msg)
            try:
                self._message_router.deliver_message(reply)
            except QMI_MessageDeliveryException:
                _logger.debug("Failed to deliver error reply to %r", message.source_address)
            except Exception:
                _logger.exception("Unexpected exception while delivering error reply to %r", message.source_address)

    def get_peer_context_names(self) -> List[str]:
        """Return a list of peer context names.

        Note that this method returns context names for outgoing connections,
        as well as local aliases for incoming connections.

        This method is thread-safe and may safely be called from any thread.
        """
        with self._lock:
            peer_context_names = list(self._peer_context_map.keys())
        return peer_context_names

    def has_peer_context(self, peer_context_name: str) -> bool:
        """Return True if a context with specified name is currently connected as a peer.

        This method is thread-safe and may safely be called from any thread.
        Note however that the result of this method may become invalid at any time
        as the set of connected contexts may change asynchronously.
        """
        with self._lock:
            return peer_context_name in self._peer_context_map


class MessageRouter:
    """Sends and delivers messages within a QMI context and between contexts.

    Each QMI context owns one `MessageRouter` instance to handle messaging.
    The `MessageRouter` provides methods to send messages and to register
    QMI objects as message receivers.

    The `MessageRouter` will create a background thread to handle network
    connections.

    This class is intended for internal use within QMI. Application programs
    should not interact with this class directly.
    """

    # Timeout (in seconds) for connecting to a peer context via TCP.
    # This can be quite short since we are working on a LAN.
    # This should be quite short, otherwise context status checking will become too slow.
    CONNECT_TIMEOUT = 2

    # Timeout (in seconds) for initial handshake on outgoing peer connections.
    HANDSHAKE_TIMEOUT = 30

    def __init__(self, context_name: str, workgroup_name: str) -> None:
        self.context_name = context_name
        self.workgroup_name = workgroup_name
        self.tcp_server_port = 0
        self._thread = None  # type: Optional[_EventDrivenThread]
        self._socket_manager = None  # type: Optional[_SocketManager]
        self._address_to_messagehandler_map = {}  # type: Dict[str, QMI_MessageHandler]
        self._address_to_messagehandler_map_lock = threading.Lock()
        self._cb_peer_context_added = None    # type: Optional[Callable[[str], None]]
        self._cb_peer_context_removed = None  # type: Optional[Callable[[str], None]]
        self._suppress_version_mismatch_warnings = False

    @property
    def suppress_version_mismatch_warnings(self) -> bool:
        """If set to `True`, no warnings will be issued when connecting to a
        peer that runs a different version of QMI.
        """

        return self._suppress_version_mismatch_warnings

    @suppress_version_mismatch_warnings.setter
    def suppress_version_mismatch_warnings(self, value: bool) -> None:

        self._suppress_version_mismatch_warnings = value

    def set_peer_context_callbacks(self,
                                   cb_peer_context_added: Optional[Callable[[str], None]],
                                   cb_peer_context_removed: Optional[Callable[[str], None]]) -> None:
        """Register callback functions to be invoked when a peer context is added or removed.

        `cb_peer_context_added(peer_context_name)` is invoked when a new outgoing
        connection to a peer context is established. This callback is not
        invoked for new incoming peer connections.

        `cb_peer_context_removed(peer_context_name)` is invoked when a peer
        connection (outgoing or incoming) is removed.

        The callbacks run in the socket manager thread. These callbacks
        are typically handled by the `SignalManager` to manage its remote
        signal subscriptions.

        This function must not be called after starting the message router.
        """
        assert self._thread is None
        self._cb_peer_context_added = cb_peer_context_added
        self._cb_peer_context_removed = cb_peer_context_removed

    def start(self) -> None:
        """Start the socket manager thread."""

        assert self._thread is None

        # Create event-driven thread for the socket manager.
        self._thread = _EventDrivenThread()
        self._thread.start()

        # Create socket manager inside the thread.
        def make_socket_manager() -> _SocketManager:
            assert self._thread is not None
            assert self._thread.event_loop is not None
            return _SocketManager(self._thread.event_loop, self)
        self._socket_manager = self._thread.run_in_thread_wait(make_socket_manager)
        assert isinstance(self._socket_manager, _SocketManager)

    def stop(self) -> None:
        """Disconnect all peer connections, stop servers and stop the socket manager thread."""

        assert self._thread is not None
        assert self._socket_manager is not None

        # Tell the socket manager to close all sockets.
        self._thread.run_in_thread(self._socket_manager.close_all)
        self._socket_manager = None

        # Stop the thread.
        self._thread.shutdown()
        self._thread.join()
        self._thread = None

    def register_message_handler(self, message_handler: QMI_MessageHandler) -> None:
        """Register a local message handler."""
        assert message_handler.address.context_id == self.context_name
        object_id = message_handler.address.object_id
        with self._address_to_messagehandler_map_lock:
            if object_id in self._address_to_messagehandler_map:
                raise QMI_DuplicateNameException("Can not register duplicate message handler address {}"
                                                 .format(message_handler.address))
            self._address_to_messagehandler_map[object_id] = message_handler

    def unregister_message_handler(self, message_handler: QMI_MessageHandler) -> None:
        """Unregister a previously registered message handler."""
        assert message_handler.address.context_id == self.context_name
        object_id = message_handler.address.object_id
        with self._address_to_messagehandler_map_lock:
            if self._address_to_messagehandler_map.get(object_id) is not message_handler:
                raise QMI_UnknownNameException("Unknown message handler {}".format(message_handler.address))
            del self._address_to_messagehandler_map[object_id]

    def start_tcp_server(self, tcp_server_port: int) -> None:
        """Start TCP server for incoming connections from remote contexts.

        Parameters:
            tcp_server_port: TCP server port, or 0 to assign a free port.
        """

        assert self._thread is not None
        assert self._socket_manager is not None
        assert self.tcp_server_port == 0

        _logger.info("Starting TCP server on port %d ...", tcp_server_port)

        sock = socket.socket(family=socket.AF_INET, type=socket.SOCK_STREAM)

        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

        # Bind the port so we can be reached from the outside.
        address = ('', tcp_server_port)  # The empty string represents 'INADDR_ANY'.
        sock.bind(address)

        # Start listening for inbound connection requests.
        # Allow for 5 pending connection requests.
        sock.listen(5)

        # Get the actual TCP port number (in case it was automatically assigned).
        (sock_host, self.tcp_server_port) = sock.getsockname()

        # Hand the server socket over to the socket manager.
        self._thread.run_in_thread_arg(self._socket_manager.add_tcp_server, sock)

    def start_udp_responder(self, udp_server_port: int) -> None:
        """Start UDP responder on the specified UDP port."""

        assert self._thread is not None
        assert self._socket_manager is not None
        assert udp_server_port != 0

        _logger.info("Starting UDP responder on port %d ...", udp_server_port)

        sock = socket.socket(family=socket.AF_INET, type=socket.SOCK_DGRAM)

        # Make sure that the same port can be re-used by multiple QMI_Contexts on the same host.
        #
        # There is a lot of similarity between the older 'SO_REUSEADDR' socket option (already present
        # in BSD before other systems copied its networking stack) and the newer 'SO_REUSEPORT'socket option.
        #
        # See https://stackoverflow.com/questions/14388706 for a deep explanation of their differences.
        #
        # Bottom line:
        #
        # - On Windows there's no choice; only SO_REUSEADDR is available;
        # - On OSX both options are available but SO_REUSEADDR doesn't work for our use-case.
        #        (we get an 'Address already in use' error).
        # - On Linux, both options are available and seem to do what we want.
        #       We default to SO_REUSEADDR.
        #
        if sys.platform in ['darwin']:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)  # type: ignore
        elif sys.platform in ['linux', 'win32']:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        else:
            sock.close()
            raise QMI_RuntimeException("Unexpected value for sys.platform ({!r})".format(sys.platform))

        # Bind the port so we can be reached from the outside.
        address = ('', udp_server_port)  # The empty string represents 'INADDR_ANY'.
        sock.bind(address)

        # Hand the UDP server socket over to the socket manager.
        self._thread.run_in_thread_arg(self._socket_manager.add_udp_responder, sock)

    def connect_to_peer(self, peer_context_name: str, peer_address: str) -> None:
        """Connect as a client to a remote QMI context at the specified peer address."""

        assert self._thread is not None
        assert self._socket_manager is not None

        # Check that the peer context name is valid (not anonymous) and not yet connected.
        if peer_context_name.startswith("$"):
            raise QMI_UsageException("Invalid peer context name {}".format(peer_context_name))
        if self._socket_manager.has_peer_context(peer_context_name):
            raise QMI_UsageException("Duplicate connection to context {} not allowed".format(peer_context_name))

        peer_addr_parsed = parse_address_and_port(peer_address)
        _logger.info("Connecting to peer context %s at %s",
                     peer_context_name,
                     format_address_and_port(peer_addr_parsed))

        # Connect to TCP server at remote context.
        outgoing_connection_socket = socket.create_connection(peer_addr_parsed, timeout=self.CONNECT_TIMEOUT)

        # Now that the connection is established, increase timeout to infinite
        # to get normal blocking semantics.
        outgoing_connection_socket.settimeout(None)

        # Disable delaying of small TCP segments.
        outgoing_connection_socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

        # Wrap socket in PeerTcpConnection object.
        conn = _PeerTcpConnection(self,
                                  outgoing_connection_socket,
                                  peer_context_name,
                                  is_incoming=False)

        # Send handshake and wait for answer.
        try:
            conn.send_handshake()
            conn.receive_handshake(self.HANDSHAKE_TIMEOUT)
        except Exception:
            conn.close()
            raise

        # Verify the name of the peer context.
        if conn.peer_context_name != peer_context_name:
            got_name = conn.peer_context_name
            conn.close()
            raise QMI_RuntimeException("Got handshake from context {} while expecting {}"
                                       .format(got_name, peer_context_name))

        # Verify the version of the peer context.
        if conn.peer_context_version != qmi.__version__ and not self._suppress_version_mismatch_warnings:
            _logger.warning("Version mismatch detected; peer context %s at %s runs version %s (local version is %s)",
                            peer_context_name, format_address_and_port(peer_addr_parsed), conn.peer_context_version,
                            qmi.__version__)

        # Hand the TCP socket over to the socket manager.
        # Note that this transfers the PeerTcpConnection object to the socket manager thread.
        self._thread.run_in_thread_wait(functools.partial(self._socket_manager.add_outgoing_connection, conn))

    def disconnect_from_peer(self, peer_context_name: str) -> None:
        """Disconnect from the specified remote QMI context.

        Raises:
            ~qmi.core.exceptions.QMI_UnknownNameException: If the specified context is not connected.
        """

        assert self._thread is not None
        assert self._socket_manager is not None

        _logger.info("Disconnecting from peer context %s", peer_context_name)
        self._thread.run_in_thread_wait(
            functools.partial(self._socket_manager.disconnect_from_peer, peer_context_name))

    def notify_peer_context_added(self, context_name: str) -> None:
        """Called by the SocketManager when an outgoing peer connection is added.

        This method runs in the socket manager thread.
        """
        if self._cb_peer_context_added is not None:
            self._cb_peer_context_added(context_name)

    def notify_peer_context_removed(self, context_name: str) -> None:
        """Called by the SocketManager when a peer connection is removed.

        This method runs in the socket manager thread.
        """
        if self._cb_peer_context_removed is not None:
            self._cb_peer_context_removed(context_name)

    def deliver_message(self, message: QMI_Message) -> None:
        """Deliver a message to a local message handler.

        This function may be called via `send_message()` for direct local
        delivery of locally originated messages.
        Alternatively, this function may be called by the socket manager
        when a message is received from a peer connection.

        This method is thread-safe. It can be called from any thread.

        Raises:
            QMI_MessageDeliveryException: If the message can not be delivered.
        """
        if message.destination_address.context_id != self.context_name:
            raise QMI_MessageDeliveryException("Can not deliver message to non-local destination {}"
                                               .format(message.destination_address))
        with self._address_to_messagehandler_map_lock:
            message_handler = self._address_to_messagehandler_map.get(message.destination_address.object_id)
        if message_handler is None:
            raise QMI_MessageDeliveryException("Can not deliver message to unknown destination {}"
                                               .format(message.destination_address))
        else:
            message_handler.handle_message(message)

    def send_message(self, message: QMI_Message) -> None:
        """Send the message to its destination.

        If the destination is within the local context, this function delivers
        the message to the local message handler. Otherwise, this function
        sends the message to the correct peer context.

        This method is thread-safe. It can be called from any thread.

        Raises:
            QMI_MessageDeliveryException: If the message can not be routed.
        """

        destination_context_name = message.destination_address.context_id
        if destination_context_name == self.context_name:

            # The destination_address is our context; perform local delivery of the message.
            self.deliver_message(message)

        else:

            # We never forward messages from remote context to remote context.
            if message.source_address.context_id != self.context_name:
                raise QMI_MessageDeliveryException(
                    "Can not send message from remote context {} to remote context {}"
                    .format(message.source_address.context_id, destination_context_name))

            # Check that the socket manager is running (message router started).
            socket_thread = self._thread
            socket_manager = self._socket_manager
            if (socket_thread is None) or (socket_manager is None):
                raise QMI_MessageDeliveryException("Can not send message to {!r} - message router inactive"
                                                   .format(message.destination_address))

            # Let the socket manager handle remote message delivery.
            if not socket_manager.has_peer_context(destination_context_name):
                raise QMI_MessageDeliveryException("Can not send message to unknown context {!r}"
                                                   .format(destination_context_name))
            socket_thread.run_in_thread_arg(socket_manager.send_message, message)

    def get_peer_context_names(self) -> List[str]:
        """Return a list of currently connected peer context names."""
        socket_manager = self._socket_manager
        if socket_manager is not None:
            return socket_manager.get_peer_context_names()
        else:
            return []



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

# We add the correct directory, enabling Sphinx to find our 'qmi_sphinx' extension.
sys.path.insert(0, os.path.abspath('.'))
sys.path.insert(0, os.path.abspath('../../..'))

# -- Project information -----------------------------------------------------

project = 'QMI'
copyright = '2019-2024, QuTech — Delft, The Netherlands'
author = 'QuTech'

# The full version, including alpha/beta/rc tags
release = '0.47.0'

# The default master_doc used to be 'index', but it was changed to 'contents'.
# Override that here (maybe rename the file to the new default later).
master_doc = 'index'
# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
#
# Order is important... autosummary should go before sphinx.ext.napoleon.
#

extensions = [ 'sphinx.ext.autodoc', 'sphinx.ext.autosummary', 'sphinx.ext.napoleon', 'sphinx.ext.todo' ]

autosummary_generate = True

autodoc_default_options = {
   # 'members'           : True,
   'member-order'      : 'bysource',
   # 'undoc_members'     : True,
   # 'private-members'   : False,
   # 'special-members'   : '',
   'inherited-members' : True,
   # 'show-inheritance'  : False,
   # 'ignore-module-all' : False,
   'imported-members'  : True
   # 'exclude-members'   : ''
}

todo_include_todos = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
#html_theme = 'alabaster'
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
#html_static_path = ['_static']



"""QMI Instrument driver for the TimeBase DIM3000 AOM driver."""

from dataclasses import dataclass, fields
from enum import IntEnum
import logging
import time
from typing import Dict, Optional, Tuple, Type, TypeVar, Union, ClassVar, get_origin, get_args
import re

from qmi.core.context import QMI_Context
from qmi.core.instrument import QMI_Instrument, QMI_InstrumentIdentification
from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.rpc import rpc_method
from qmi.core.transport import create_transport


# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


T = TypeVar("T", bound="DIM3000Base")


class DIM3000SweepMode(IntEnum):
    """Options for frequency sweep mode."""

    OFF = 0
    TRI_INT_TRIG = 1
    TRI_EXT_TRIG = 2
    SAW_INT_TRIG = 3
    SAW_EXT_TRIG = 4


class DIM3000FMDeviation(IntEnum):
    """Options for available frequency deviation during frequency modulation."""

    _3200HZ = 0
    _6400HZ = 1
    _12800HZ = 2
    _25600HZ = 3
    _51200HZ = 4
    _102400HZ = 5
    _204800HZ = 6
    _409600HZ = 7
    _819200HZ = 8
    _1638400HZ = 9
    _3276800HZ = 10
    _6553600HZ = 11
    _13107200HZ = 12
    _26214400HZ = 13
    _52428800HZ = 14
    _104857600HZ = 15


@dataclass(frozen=True)
class DIM3000Base:
    """Base class that gives all dataclasses the factory function."""

    PATTERN: ClassVar[str]
    """RegEx that will be used for pattern matching on the string received from the device."""

    @classmethod
    def from_string(cls: Type[T], string: str) -> T:
        """Factory function for making a dataclass from a string received from device."""

        if (match := re.match(cls.PATTERN, string)) is None:
            raise QMI_InstrumentException(f"Unexpected string returned from device: {string}")

        kwargs: Dict[str, Union[int, float, bool, str, None]] = {}

        for field in fields(cls):
            # remap optional field type
            if get_origin(field.type) is Union:
                field_type = get_args(field.type)[0]
            else:
                field_type = field.type

            # optional fields might not have match, so set to None
            if match[field.name] is None:
                kwargs[field.name] = None
            elif field_type is int:
                kwargs[field.name] = int(match[field.name])
            elif field_type is float:
                kwargs[field.name] = float(match[field.name])
            elif field_type is bool:
                kwargs[field.name] = bool(int(match[field.name]))
            elif field_type is str:
                kwargs[field.name] = match[field.name]
            elif issubclass(field_type, IntEnum):
                kwargs[field.name] = field_type(int(match[field.name]))
            else:
                raise RuntimeError(f"Unexpected type {field_type} for {cls}!")

        return cls(**kwargs)


@dataclass(frozen=True)
class DIM3000DevInfo(DIM3000Base):
    """Dataclass containing device information."""

    PATTERN = r"^Rdev:(?P<dev>\w*)\|" r"Rhv:(?P<hv>\w*)\|" r"Rfv:(?P<fv>\w*)\|" r"Rfb:(?P<fb>\w*)\|" r"Rsn:(?P<sn>\w*)"
    dev: str
    hv: str
    fv: str
    fb: str
    sn: str


@dataclass(frozen=True)
class DIM3000InitData(DIM3000Base):
    """Dataclass containing initial data."""

    PATTERN = (
        r"^Ramoffsmin:(?P<amoffsmin>-?\d*)\|"
        r"Ramoffsmax:(?P<amoffsmax>-?\d*)\|"
        r"Ramoffsnom:(?P<amoffsnom>-?\d*)\|"
        r"(Rbtstat:(?P<btstat>\d)\|)?"
        r"(Radcoffs:(?P<adcoffs>-?\d*)\|)?"
        r"Rinit:(?P<init>\d)"
    )
    amoffsmin: int
    amoffsmax: int
    amoffsnom: int
    btstat: Optional[bool]
    adcoffs: Optional[int]
    init: bool


@dataclass(frozen=True)
class DIM3000Parameters(DIM3000Base):
    """Dataclass containing parameter data."""

    PATTERN = (
        r"^Rfreq:(?P<freq>\d*)\|"
        r"Rampl:(?P<ampl>\d*)\|"
        r"Rout:(?P<out>\d*)\|"
        r"Rpmon:\d*\|"
        r"Rpmfr:\d*\|"
        r"Rpmd:\d*\|"
        r"Rpmphc:\d*\|"
        r"Rswpm:(?P<swpm>\d*)\|"
        r"Rswps:(?P<swps>\d*)\|"
        r"Rswpp:(?P<swpp>\d*)\|"
        r"Rswpf:(?P<swpf>\d*)\|"
        r"Rswpt:(?P<swpt>\d*)\|"
        r"Rfmon:(?P<fmon>\d*)\|"
        r"Rfmdev:(?P<fmdev>\d*)\|"
        r"Rplson:(?P<plson>\d*)\|"
        r"Rplsfr:(?P<plsfr>\d*)\|"
        r"Rplsdt:(?P<plsdt>\d*)\|"
        r"Rffreq:(?P<ffreq>\d*)\|"
        r"Rfampl:(?P<fampl>\d*)\|"
        r"Ramoffs:(?P<amoffs>-?\d*)\|"
        r"Rpcbtemp:(?P<pcbtemp>\d*)\|"
        r"Rrefstat:(?P<refstat>\d*)\|"
        r"Rreflev:(?P<reflev>-?\d*)\|"
        r"Rvcclev:(?P<vcclev>\d*)"
    )
    freq: int
    ampl: float
    out: bool
    swpm: DIM3000SweepMode
    swps: int
    swpp: int
    swpf: int
    swpt: int
    fmon: bool
    fmdev: DIM3000FMDeviation
    plson: bool
    plsfr: int
    plsdt: int
    ffreq: int
    fampl: float
    amoffs: int
    pcbtemp: float
    refstat: bool
    reflev: int
    vcclev: float

    def __post_init__(self):
        # Correct some scaling of parameters
        # Using __setattr__ because this is what is recommended by python when working with frozen classes
        # See https://docs.python.org/3/library/dataclasses.html#frozen-instances
        object.__setattr__(self, "ampl", self.ampl / 10.0)
        object.__setattr__(self, "fampl", self.fampl / 10.0)
        object.__setattr__(self, "pcbtemp", self.pcbtemp / 100.0)
        object.__setattr__(self, "vcclev", self.vcclev / 100.0)


class TimeBase_DIM3000(QMI_Instrument):
    """QMI Instrument driver for the TimeBase DIM3000 AOM driver."""

    _rpc_constants = [
        "FREQ_RANGE",
        "TIME_RANGE",
        "PULSE_FREQ_RANGE",
        "DUTY_CYCLE_RANGE",
        "AM_OFFSET_RANGE",
        "MINIMUM_EXEC_DELAY_S",
    ]

    # Public class constants
    FREQ_RANGE = (10, 400_000_000)
    TIME_RANGE = (4, 262000)
    PULSE_FREQ_RANGE = (20, 1000)
    DUTY_CYCLE_RANGE = (1, 99)
    AM_OFFSET_RANGE = (-255, 25)
    MINIMUM_EXEC_DELAY_S = 0.1  # Maximum execution speed of 10 cmds/sec from manual

    _RESPONSE_TIMEOUT_S = 2.0
    _TERMINATOR = b"\n"

    def __init__(self, context: QMI_Context, name: str, transport: str) -> None:
        """Initialization of the TimeBase DIM3000 instrument driver.

        Note:
            Maximum execution speed is about 10 commands/sec.

        Parameters:
            name: Name for this instrument instance.
            transport: Transport descriptor to access the instrument.
        """
        super().__init__(context, name)
        self._transport = create_transport(transport, default_attributes={"baudrate": 19200})
        self._prev_cmd_ts = 0.0

    def _write_par(self, par: str, val: int) -> None:
        """Set a parameter of the device with a specified value."""
        self._check_is_open()
        self._check_cmd_exec_speed(time.monotonic())
        self._transport.write(f"S{par}:{val}".encode("ascii") + self._TERMINATOR)

    def _ask(self, cmd: str) -> str:
        """Send command to instrument and return response from instrument."""
        self._check_is_open()
        self._check_cmd_exec_speed(time.monotonic())
        self._transport.discard_read()
        self._transport.write(cmd.encode("ascii") + self._TERMINATOR)
        resp = self._transport.read_until(message_terminator=self._TERMINATOR, timeout=self._RESPONSE_TIMEOUT_S)
        return resp.rstrip().decode("ascii")

    def _check_cmd_exec_speed(self, now: float) -> None:
        """Checks if command execution is higher than expected."""
        if (delay := now - self._prev_cmd_ts) < self.MINIMUM_EXEC_DELAY_S:
            _logger.warning(
                "Delay since previous command (%f) is shorter than recommended (%f)!", delay, self.MINIMUM_EXEC_DELAY_S
            )
        self._prev_cmd_ts = now

    def _check_range(self, range: Tuple[int, int], val: int) -> None:
        """Checks if value is in range."""
        if not range[0] <= val <= range[1]:
            raise ValueError(f"Value ({val}) is not in range {range}!")

    @rpc_method
    def open(self) -> None:
        _logger.info("[%s] Opening connection to instrument", self._name)
        self._check_is_closed()
        self._transport.open()
        super().open()

    @rpc_method
    def close(self) -> None:
        _logger.info("[%s] Closing connection to instrument", self._name)
        super().close()
        self._transport.close()

    @rpc_method
    def get_idn(self) -> QMI_InstrumentIdentification:
        """Get the device info in the `QMI_InstrumentIdentification` format."""
        dev_info = self.get_device_info()
        return QMI_InstrumentIdentification(
            vendor="TimeBase",
            model=f"DIM3000[{dev_info.dev}]",
            serial=dev_info.sn,
            version=f"{dev_info.hv}.{dev_info.fv}.{dev_info.fb}",
        )

    @rpc_method
    def get_device_info(self) -> DIM3000DevInfo:
        """Get the device info from the device."""
        return DIM3000DevInfo.from_string(self._ask("Gdev"))

    @rpc_method
    def get_init_data(self) -> DIM3000InitData:
        """Get initial data from the device."""
        return DIM3000InitData.from_string(self._ask("Ginit"))

    @rpc_method
    def get_parameters(self) -> DIM3000Parameters:
        """Get parameters from the device."""
        return DIM3000Parameters.from_string(self._ask("Gpar"))

    @rpc_method
    def set_output_frequency(self, freq_hz: int) -> None:
        """Set the output frequency in Hz."""
        self._check_range(self.FREQ_RANGE, freq_hz)
        self._write_par("freq", int(freq_hz))

    @rpc_method
    def set_output_amplitude(self, ampl_dbm: float) -> None:
        """Set the output amplitude in dBm."""
        self._write_par("ampl", int(ampl_dbm * 10.0))

    @rpc_method
    def set_sweep_mode(self, swpm: DIM3000SweepMode) -> None:
        """Set the sweep mode. See `DIM3000SweepMode`."""
        self._write_par("swpm", swpm.value)

    @rpc_method
    def set_sweep_start_frequency(self, swps_hz: int) -> None:
        """Set the sweep start frequency in Hz."""
        self._check_range(self.FREQ_RANGE, swps_hz)
        self._write_par("swps", int(swps_hz))

    @rpc_method
    def set_sweep_stop_frequency(self, swpp_hz: int) -> None:
        """Set the sweep stop frequency in Hz."""
        self._check_range(self.FREQ_RANGE, swpp_hz)
        self._write_par("swpp", int(swpp_hz))

    @rpc_method
    def set_sweep_step_frequency(self, swpf_hz: int) -> None:
        """Set the sweep step frequency in Hz."""
        self._check_range(self.FREQ_RANGE, swpf_hz)
        self._write_par("swpf", int(swpf_hz))

    @rpc_method
    def set_sweep_step_time(self, swpt_ns: int) -> None:
        """Set the sweep step time in nanoseconds."""
        self._check_range(self.TIME_RANGE, swpt_ns)
        self._write_par("swpt", int(swpt_ns))

    @rpc_method
    def set_fm_input(self, fmon: bool) -> None:
        """Set FM input."""
        self._write_par("fmon", int(fmon))

    @rpc_method
    def enable_fm_input(self) -> None:
        """Enable FM input."""
        self._write_par("fmon", int(True))

    @rpc_method
    def disable_fm_input(self) -> None:
        """Disable FM input."""
        self._write_par("fmon", int(False))

    @rpc_method
    def set_fm_deviation(self, fmdev: DIM3000FMDeviation) -> None:
        """Set FM deviation. See `DIM3000FMDeviation`."""
        self._write_par("fmdev", fmdev.value)

    @rpc_method
    def set_pulse_mode(self, plson: bool) -> None:
        """Set pulse mode."""
        self._write_par("plson", int(plson))

    @rpc_method
    def enable_pulse_mode(self) -> None:
        """Enable pulse mode."""
        self._write_par("plson", int(True))

    @rpc_method
    def disable_pulse_mode(self) -> None:
        """Disable pulse mode."""
        self._write_par("plson", int(False))

    @rpc_method
    def set_pulse_frequency(self, plsfr_hz: int) -> None:
        """Set the pulse frequency in Hz."""
        self._check_range(self.PULSE_FREQ_RANGE, plsfr_hz)
        self._write_par("plsfr", plsfr_hz)

    @rpc_method
    def set_pulse_duty_cycle(self, plsdt: int) -> None:
        """Set the pulse duty cycle [1-99]."""
        self._check_range(self.DUTY_CYCLE_RANGE, plsdt)
        self._write_par("plsdt", plsdt)

    @rpc_method
    def set_fsk_frequency(self, ffreq_hz: int) -> None:
        """Set the FSK frequency in Hz."""
        self._check_range(self.FREQ_RANGE, ffreq_hz)
        self._write_par("ffreq", ffreq_hz)

    @rpc_method
    def set_fsk_amplitude(self, fampl_dbm: float) -> None:
        """Set the FSK amplitude in dBm."""
        self._write_par("fampl", int(fampl_dbm * 10.0))

    @rpc_method
    def set_am_offset(self, amoffs: int) -> None:
        """Set the AM offset."""
        self._check_range(self.AM_OFFSET_RANGE, amoffs)
        self._write_par("amoffs", amoffs)



import logging
import unittest
import unittest.mock

import numpy as np
import pydwf
import pydwf.utilities
from pydwf.core.dwf_device import DwfDevice

import qmi
from qmi.core.exceptions import QMI_InvalidOperationException, QMI_InstrumentException, QMI_UsageException, \
    QMI_TimeoutException
from qmi.instruments.digilent import Digilent_AnalogDiscovery2, OnClose, Filter


class AnalogDiscovery2OpenCloseTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self._device_mock = unittest.mock.MagicMock(spec=DwfDevice)

        patcher = unittest.mock.patch('pydwf.DwfLibrary')
        self._dwf_lib_mock = patcher.start().return_value

        open_patcher = unittest.mock.patch('pydwf.utilities.openDwfDevice', return_value=self._device_mock)
        self._open_dwf_device_mock = open_patcher.start()

        self.addCleanup(patcher.stop)
        self.addCleanup(open_patcher.stop)

        qmi.start('analog_discovery_openclose_test')

        logging.getLogger("qmi.core.instrument").setLevel(logging.CRITICAL)
        self._ad2: Digilent_AnalogDiscovery2 = qmi.make_instrument('AD2', Digilent_AnalogDiscovery2, '210321AD12A4')

    def tearDown(self) -> None:
        qmi.stop()
        logging.getLogger("qmi.core.instrument").setLevel(logging.NOTSET)

    def test_open_close(self):

        self._ad2.open()
        self._ad2.close()

        self._open_dwf_device_mock.assert_called_once_with(self._dwf_lib_mock, serial_number_filter='210321AD12A4')

    def test_open_fail_when_not_found(self):

        def raise_value_error(lib, serial_number_filter):
            raise ValueError('Fake value error')

        self._open_dwf_device_mock.side_effect = raise_value_error

        with self.assertRaises(ValueError):
            self._ad2.open()

    def test_open_raises_instrument_exception_when_lib_error(self):

        def raise_value_error(lib, serial_number_filter):
            raise pydwf.DwfLibraryError(code=pydwf.DwfErrorCode.UnknownError, msg='Fake lib error')

        self._open_dwf_device_mock.side_effect = raise_value_error

        with self.assertRaises(QMI_InstrumentException):
            self._ad2.open()

    def test_open_fails_when_opened_double(self):

        self._ad2.open()
        with self.assertRaises(QMI_InvalidOperationException):
            self._ad2.open()
        self._ad2.close()

    def test_close_fails_when_closed_before_opened(self):

        with self.assertRaises(QMI_InvalidOperationException):
            self._ad2.close()

    def test_raise_usage_exception_when_invoke_method_without_open(self):

        with self.assertRaises(QMI_UsageException):
            self._ad2.prepare_analog_output_channel_for_static_output(1, 1.5)

        with self.assertRaises(QMI_UsageException):
            self._ad2.set_analog_output_voltage_output(1, 1.5)

        with self.assertRaises(QMI_UsageException):
            self._ad2.get_analog_output_voltage_output(1)

        with self.assertRaises(QMI_UsageException):
            self._ad2.set_device_on_close(OnClose.SHUTDOWN)

        with self.assertRaises(QMI_UsageException):
            self._ad2.reset_analog_input()

        with self.assertRaises(QMI_UsageException):
            self._ad2.prepare_analog_input_sample(1, 10.0, 0.001)

        with self.assertRaises(QMI_UsageException):
            self._ad2.get_analog_input_sample(1)

        with self.assertRaises(QMI_UsageException):
            self._ad2.get_analog_input_acquire_samples(1)

        with self.assertRaises(QMI_UsageException):
            self._ad2.prepare_analog_input_record(1, 10.0, 1.0, 0.001)

        with self.assertRaises(QMI_UsageException):
            self._ad2.get_analog_input_record(1)


class AnalogDiscovery2ErrorTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self._device_mock = unittest.mock.MagicMock(spec=DwfDevice)

        patcher = unittest.mock.patch('pydwf.DwfLibrary')
        self._dwf_lib_mock = patcher.start().return_value

        open_patcher = unittest.mock.patch('pydwf.utilities.openDwfDevice', return_value=self._device_mock)
        self._open_dwf_device_mock = open_patcher.start()

        self.addCleanup(patcher.stop)
        self.addCleanup(open_patcher.stop)

        qmi.start('analogdiscovery2_error_test')

    def tearDown(self) -> None:
        qmi.stop()
        logging.getLogger("qmi.core.instrument").setLevel(logging.NOTSET)

    def test_close_raises_instrument_error_when_lib_error(self):
        def raise_value_error():
            raise pydwf.DwfLibraryError(code=pydwf.DwfErrorCode.UnknownError, msg='Fake lib error')
        self._device_mock.close.side_effect = raise_value_error

        ad2: Digilent_AnalogDiscovery2 = qmi.make_instrument('AD2', Digilent_AnalogDiscovery2, '210321AD12A4')

        ad2.open()

        with self.assertRaises(QMI_InstrumentException):
            ad2.close()

        # Really close now to avoid ResourceWarning.
        self._device_mock.close.side_effect = None
        ad2.close()


class AnalogDiscovery2MethodsTestCase(unittest.TestCase):

    def setUp(self) -> None:

        self._device_mock = unittest.mock.MagicMock(spec=DwfDevice)

        patcher = unittest.mock.patch('pydwf.DwfLibrary')
        self._dwf_lib_mock = patcher.start().return_value

        open_patcher = unittest.mock.patch('pydwf.utilities.openDwfDevice', return_value=self._device_mock)
        self._open_dwf_device_mock = open_patcher.start()

        self.addCleanup(patcher.stop)
        self.addCleanup(open_patcher.stop)

        qmi.start('analogdiscovery2_methods_test')
        self._ad2: Digilent_AnalogDiscovery2 = qmi.make_instrument('AD2', Digilent_AnalogDiscovery2, '210321AD12A4')

        self._analog_out_mock = unittest.mock.create_autospec(spec=pydwf.core.dwf_device.AnalogOut)
        self._device_mock.analogOut = self._analog_out_mock

        self._analog_in_mock = unittest.mock.create_autospec(spec=pydwf.core.dwf_device.AnalogIn)
        self._device_mock.analogIn = self._analog_in_mock

        self._ad2.open()

    def tearDown(self) -> None:
        self._ad2.close()
        qmi.stop()

    def test_prepare_analog_output_channel_for_static_output(self):
        self._ad2.prepare_analog_output_channel_for_static_output(1, 5.5)

        self._analog_out_mock.nodeFunctionSet.assert_called_with(
            1, pydwf.DwfAnalogOutNode.Carrier, pydwf.DwfAnalogOutFunction.Square
        )
        self._analog_out_mock.idleSet.assert_called_with(1, pydwf.DwfAnalogOutIdle.Initial)
        self._analog_out_mock.nodeAmplitudeSet.assert_called_with(1,  pydwf.DwfAnalogOutNode.Carrier, 5.5)
        self._analog_out_mock.nodeEnableSet.assert_called_with(1,  pydwf.DwfAnalogOutNode.Carrier, True)

    def raise_digilent_waveform_library_error(self, *args):
        raise pydwf.DwfLibraryError(code=pydwf.DwfErrorCode.UnknownError, msg='Fake lib error')

    def test_prepare_analog_output_channel_for_static_output_raises_qmi_instrument_error(self):
        self._analog_out_mock.nodeFunctionSet.side_effect = self.raise_digilent_waveform_library_error
        with self.assertRaises(QMI_InstrumentException):
            self._ad2.prepare_analog_output_channel_for_static_output(1, 1.0)

    def test_set_analog_output_voltage_output(self):
        self._ad2.set_analog_output_voltage_output(1, 1.5)
        self._analog_out_mock.nodeAmplitudeSet.assert_called_with(1, pydwf.DwfAnalogOutNode.Carrier, 1.5)

    def test_set_analog_output_voltage_output_raises_qmi_instrument_error(self):
        self._analog_out_mock.nodeAmplitudeSet.side_effect = self.raise_digilent_waveform_library_error
        with self.assertRaises(QMI_InstrumentException):
            self._ad2.set_analog_output_voltage_output(1, 1.0)

    def test_get_analog_output_voltage_output(self):
        expected_voltage = 1.5
        self._analog_out_mock.nodeAmplitudeGet.return_value = expected_voltage

        voltage = self._ad2.get_analog_output_voltage_output(1)

        self._analog_out_mock.nodeAmplitudeGet.assert_called_with(1, pydwf.DwfAnalogOutNode.Carrier)

        self.assertEqual(voltage, expected_voltage)

    def test_get_analog_output_voltage_output_raises_qmi_instrument_error(self):
        self._analog_out_mock.nodeAmplitudeGet.side_effect = self.raise_digilent_waveform_library_error
        with self.assertRaises(QMI_InstrumentException):
            self._ad2.get_analog_output_voltage_output(1)

    def test_set_device_on_close(self):
        self._ad2.set_device_on_close(OnClose.SHUTDOWN)
        self._device_mock.paramSet.assert_called_with(pydwf.DwfDeviceParameter.OnClose, OnClose.SHUTDOWN.value)

    def test_set_device_on_close_raises_qmi_instrument_error(self):
        self._device_mock.paramSet.side_effect = self.raise_digilent_waveform_library_error
        with self.assertRaises(QMI_InstrumentException):
            self._ad2.set_device_on_close(OnClose.SHUTDOWN)

    def test_reset_analog_input(self):
        self._ad2.reset_analog_input()
        self._analog_in_mock.reset.assert_called()

    def test_reset_analog_input_raises_qmi_instrument_error(self):
        self._analog_in_mock.reset.side_effect = self.raise_digilent_waveform_library_error
        with self.assertRaises(QMI_InstrumentException):
            self._ad2.reset_analog_input()

    def test_prepare_analog_input_sample(self):
        expected_channel, expected_range, expected_freq = (1, 10.0, 0.001)
        self._ad2.prepare_analog_input_sample(expected_channel, expected_range, expected_freq)

        self._analog_in_mock.channelEnableSet.assert_called_with(expected_channel, True)
        self._analog_in_mock.frequencySet.assert_called_with(expected_freq)

        self._analog_in_mock.channelOffsetSet.assert_called_with(expected_channel, 0.0)
        self._analog_in_mock.channelRangeSet.assert_called_with(expected_channel, expected_range)
        self._analog_in_mock.channelFilterSet.assert_called_with(expected_channel, pydwf.DwfAnalogInFilter(Filter.AVERAGE.value))
        self._analog_in_mock.configure.assert_called_with(False, False)

    def test_prepare_analog_input_sample_with_buffer_size(self):
        expected_channel, expected_range, expected_freq = (1, 10.0, 0.001)
        self._ad2.prepare_analog_input_sample(expected_channel, expected_range, expected_freq,
                                              voltage_offset=0.0,
                                              buffer_size=1024)

        self._analog_in_mock.channelEnableSet.assert_called_with(expected_channel, True)
        self._analog_in_mock.frequencySet.assert_called_with(expected_freq)

        self._analog_in_mock.channelOffsetSet.assert_called_with(expected_channel, 0.0)
        self._analog_in_mock.channelRangeSet.assert_called_with(expected_channel, expected_range)
        self._analog_in_mock.channelFilterSet.assert_called_with(expected_channel, pydwf.DwfAnalogInFilter(Filter.AVERAGE.value))
        self._analog_in_mock.configure.assert_called_with(False, False)
        self._analog_in_mock.bufferSizeSet.assert_called_with(1024)

    def test_prepare_analog_input_sample_raises_qmi_instrument_error(self):
        self._analog_in_mock.channelEnableSet.side_effect = self.raise_digilent_waveform_library_error
        with self.assertRaises(QMI_InstrumentException):
            expected_channel, expected_range, expected_freq = (1, 10.0, 0.001)
            self._ad2.prepare_analog_input_sample(expected_channel, expected_range, expected_freq)

    def test_get_analog_input_sample(self):
        expected_sample_value = 6.62607
        self._analog_in_mock.statusSample.return_value = expected_sample_value
        sample = self._ad2.get_analog_input_sample(1)

        self._analog_in_mock.status.assert_called_with(False)
        self._analog_in_mock.statusSample.assert_called_with(1)
        self.assertEqual(sample, expected_sample_value)

    def test_get_analog_input_sample_raises_qmi_instrument_error(self):
        self._analog_in_mock.status.side_effect = self.raise_digilent_waveform_library_error
        with self.assertRaises(QMI_InstrumentException):
            self._ad2.get_analog_input_sample(1)

    def test_get_analog_input_acquire_samples(self):
        expected_buffer_size = 1024
        expected_samples = np.ones(expected_buffer_size, dtype=np.float64)
        self._analog_in_mock.status.side_effect = [pydwf.DwfState.Running, pydwf.DwfState.Done]
        self._analog_in_mock.bufferSizeGet.return_value = expected_buffer_size
        self._analog_in_mock.statusData.return_value = expected_samples

        samples = self._ad2.get_analog_input_acquire_samples(1)

        self._analog_in_mock.configure.assert_called_with(True, True)
        self._analog_in_mock.status.assert_called_with(True)
        self._analog_in_mock.bufferSizeGet.assert_called()
        self._analog_in_mock.statusData.assert_called_with(1, expected_buffer_size)

        np.testing.assert_array_equal(expected_samples, samples)

    def test_get_analog_input_acquire_samples_raises_qmi_instrument_error(self):
        self._analog_in_mock.configure.side_effect = self.raise_digilent_waveform_library_error
        with self.assertRaises(QMI_InstrumentException):
            self._ad2.get_analog_input_acquire_samples(1)

    def test_get_analog_input_acquire_samples_raises_qmi_timeout_exceptions(self):
        with self.assertRaises(QMI_TimeoutException):
            self._ad2.get_analog_input_acquire_samples(1, 0.0000000001)

    def test_prepare_analog_input_record(self):
        self._ad2.prepare_analog_input_record(1, 10.0, 1.0, 0.001)

        self._analog_in_mock.channelEnableSet.assert_called_with(1, True)
        self._analog_in_mock.channelOffsetSet.assert_called_with(1, 0.0)
        self._analog_in_mock.channelRangeSet.assert_called_with(1, 10.0)
        self._analog_in_mock.acquisitionModeSet.assert_called_with(pydwf.DwfAcquisitionMode.Record)
        self._analog_in_mock.frequencySet.assert_called_with(0.001)
        self._analog_in_mock.recordLengthSet.assert_called_with(1.0)
        self._analog_in_mock.channelFilterSet.assert_called_with(1, pydwf.DwfAnalogInFilter(Filter.AVERAGE.value))

    def test_prepare_analog_input_record_raises_qmi_instrument_error(self):
        self._analog_in_mock.channelEnableSet.side_effect = self.raise_digilent_waveform_library_error
        with self.assertRaises(QMI_InstrumentException):
            self._ad2.prepare_analog_input_record(1, 10.0, 1.0, 0.001)

    def test_get_analog_input_record(self):
        self._analog_in_mock.status.return_value = pydwf.DwfState.Running
        self._analog_in_mock.statusRecord.return_value = (5, 0, 0)
        self._analog_in_mock.statusData.return_value = np.ones(5, dtype=np.float64)

        sample_record = self._ad2.get_analog_input_record(1, 10)

        np.testing.assert_array_equal(np.ones(10, dtype=np.float64), sample_record)

    def test_get_analog_input_record_advanced(self):
        self._analog_in_mock.status.side_effect = [pydwf.DwfState.Config, pydwf.DwfState.Prefill,
                                                   pydwf.DwfState.Armed, pydwf.DwfState.Running,
                                                   pydwf.DwfState.Running, pydwf.DwfState.Running]
        self._analog_in_mock.statusRecord.side_effect = [(0, 0, 0), (5, 0, 0), (6, 0, 0)]
        self._analog_in_mock.statusData.return_value = np.ones(5, dtype=np.float64)
        self._analog_in_mock.recordLengthGet.return_value = 10
        self._analog_in_mock.frequencyGet.return_value = 1.0

        sample_record = self._ad2.get_analog_input_record(1)

        np.testing.assert_array_equal(np.ones(10, dtype=np.float64), sample_record)

    def test_get_analog_input_record_raises_qmi_instrument_error(self):
        self._analog_in_mock.configure.side_effect = self.raise_digilent_waveform_library_error
        with self.assertRaises(QMI_InstrumentException):
            self._ad2.get_analog_input_record(1, 10)

    def test_get_analog_input_record_data_lost_raises_instrument_error(self):
        self._analog_in_mock.status.return_value = pydwf.DwfState.Running
        self._analog_in_mock.statusRecord.return_value = (5, 1, 0)
        self._analog_in_mock.statusData.return_value = np.ones(5, dtype=np.float64)

        with self.assertRaises(QMI_InstrumentException):
            self._ad2.get_analog_input_record(1, 10)

    def test_get_analog_input_record_corrupt_raises_instrument_error(self):
        self._analog_in_mock.status.return_value = pydwf.DwfState.Running
        self._analog_in_mock.statusRecord.return_value = (5, 0, 1)
        self._analog_in_mock.statusData.return_value = np.ones(5, dtype=np.float64)

        with self.assertRaises(QMI_InstrumentException):
             self._ad2.get_analog_input_record(1, 10)


if __name__ == '__main__':
    unittest.main()



import enum
import time
from threading import Condition, Lock
from typing import Callable, Dict, List, Tuple

import numpy as np

from qmi.core.exceptions import QMI_InvalidOperationException, QMI_RuntimeException, QMI_UsageException
from qmi.core.thread import QMI_Thread
from qmi.instruments.picoquant import EventDataType, EventFilterMode, SYNC_TYPE
from qmi.instruments.picoquant.support._decoders import EventDecoder, _T2EventDecoder, _T3EventDecoder
from qmi.instruments.picoquant.support._realtime import _RealTimeHistogramProcessor, _get_sync_deltas


@enum.unique
class _MODE(enum.Enum):
    """Symbolic constants for the :func:`~MultiHarpDevice.initialize` method's `mode` argument.

    These are defined as preprocessor symbols in the ``mhdefin.h`` C header file.
    """
    HIST = 0
    """Histogram mode."""
    T2 = 2
    """T2 mode."""
    T3 = 3
    """T3 mode."""
    CONT = 8
    """Continuous Mode"""


class _FetchEventsThread(QMI_Thread):
    """This thread continuously fetches event data from the MultiHarp via USB."""

    _LOOP_SLEEP_DURATION = 0.01

    def __init__(self,
                 read_fifo_func: Callable[[], np.ndarray],
                 publish_histogram_func: Callable,
                 publish_countrate_func: Callable,
                 max_pending_events: int = 10**8
                 ) -> None:
        """Initialize background event fetching thread.

        Parameters:
            read_fifo_func: Callable that can be used to obtain an ndarray of FIFO data.
            publish_histogram_func: Callable used to publish histogram data.
            publish_countrate_func: callable used to countrate histogram data.
            max_pending_events: The purpose of this limit is to avoid consuming an excessive amount of memory.
                By default allow at most 10**8 events in the queue (~ 900 MByte). It is not expected that this limit
                will be exceeded in a properly working setup.

        """
        super().__init__()
        self._read_fifo_func = read_fifo_func
        self._publish_histogram_func = publish_histogram_func
        self._publish_countrate_func = publish_countrate_func
        self._mode = _MODE.HIST  # Start up _mode class variable in HIST mode to give a value. Will be T2 or T3 later.
        self._condition = Condition(Lock())
        self._decoder = EventDecoder()
        self._event_filter = _EventFilter()
        self._histogram_processor = _RealTimeHistogramProcessor(publish_histogram_func, publish_countrate_func)
        self._event_filter_channels: Dict[int, EventFilterMode] = {}
        self._event_filter_aperture = (0, 0)
        self._histogram_channels: List[int] = []
        self._histogram_resolution = 1
        self._histogram_num_bins = 0
        self._histogram_num_sync = 0
        self._countrate_aperture = (0, 0)
        self._countrate_num_sync = 0
        self._active = False
        self._count_read_fifo = 0
        self._block_events = False
        self._data: List[np.ndarray] = []
        self._data_timestamp = 0.0
        self._num_pending_events = 0
        self._pending_events_overflow = False
        self._max_pending_events = max_pending_events

    def _process_fifo_data(self, fifo_data: np.ndarray, fifo_data_timestamp: float) -> None:
        """Process new received raw event records from the multiharp, hydraharp.

        This function will be called in the background 'run' thread while holding the "self._condition" lock.
        """
        # Decode event records.
        event_records = self._decoder.process_data(fifo_data)
        # Update real-time histogram
        self._histogram_processor.process_events(event_records)

        if not self._block_events:

            event_records = self._event_filter.process_events(event_records)

            # Store filtered events.
            num_events = len(event_records)
            if num_events > 0:
                if self._num_pending_events + num_events > self._max_pending_events:
                    self._pending_events_overflow = True
                else:
                    self._data.append(event_records)
                    self._num_pending_events += num_events
                    self._data_timestamp = fifo_data_timestamp

    def run(self) -> None:
        """Main function running in the background thread."""
        with self._condition:
            while not self._shutdown_requested:
                if self._active:
                    # Read data from a Harp, if any.
                    fifo_data = self._read_fifo_func()
                    fifo_data_timestamp = time.time()
                    if len(fifo_data) > 0:
                        # Process received event records.
                        self._process_fifo_data(fifo_data, fifo_data_timestamp)

                    self._count_read_fifo += 1
                    # Sleep and wake up again in 10ms.
                    self._condition.wait(self._LOOP_SLEEP_DURATION)

                else:
                    # Wait until we are activated.
                    self._condition.wait()

    def activate(self, mode: _MODE = _MODE.T2, sync_frequency_hz: float = 5E6, resolution_ps: float = 1.0) -> None:
        """Activate background event fetching.

        This method is called immediately after starting a measurement in TTTR mode.
        It activates continuous event fetching in the background thread.

        This method is thread-safe.
        It will be called in the thread that owns the `PicoQuant_MultiHarp150`, `PicoQuant_HydraHarp400` instance.

        Parameters:
            mode: The mode to activate the event fetching in. Can be _MODE.T2 (default) or _MODE.T3.
            sync_frequency_hz: The synchronization signal frequency (in Hz). Applicable only in T3 mode. Default 5 MHz.
            resolution_ps: The resolution of a 15-bit d_time bit in picoseconds. Default is 1.0 ps.
        """
        self._mode = mode
        with self._condition:
            if self._active:
                raise QMI_InvalidOperationException("Already active")

            if self._mode == _MODE.T2:
                self._decoder = _T2EventDecoder()

            elif self._mode == _MODE.T3:
                self._decoder = _T3EventDecoder(sync_frequency_hz, resolution_ps)

            else:
                raise QMI_UsageException(f"No event processing decoder for mode {self._mode}")

            self._event_filter.set_event_filter_config(self._event_filter_channels,
                                                       self._event_filter_aperture)
            self._histogram_processor.set_histogram_config(self._histogram_channels,
                                                           self._histogram_resolution,
                                                           self._histogram_num_bins,
                                                           self._histogram_num_sync)
            self._histogram_processor.set_countrate_config(self._countrate_aperture,
                                                           self._countrate_num_sync)
            self._count_read_fifo = 0
            self._active = True
            self._data = []
            self._num_pending_events = 0
            self._pending_events_overflow = False
            self._condition.notify_all()

    def deactivate(self) -> None:
        """Deactivate background event fetching.

        This method is called after stopping a measurement in TTTR mode.
        Note that the instrument typically continues to produce events for a
        short time after stopping a measurement.

        This method is thread-safe.
        It will be called in the thread that owns the `PicoQuant_MultiHarp150`, `PicoQuant_HydraHarp400` instance.
        """
        with self._condition:
            self._active = False
            self._condition.notify_all()

    def set_block_events(self, blocked: bool) -> None:
        """Enable or disable blocking of events.

        This method is thread-safe.
        It will be called in the thread that owns the `PicoQuant_MultiHarp150`, `PicoQuant_HydraHarp400` instance.

        Parameters:
            blocked: A boolean flag to enable or disable event block.

        """
        with self._condition:
            self._block_events = blocked

    def set_event_filter_config(self,
                                channel_filter: Dict[int, EventFilterMode],
                                sync_aperture: Tuple[int, int]
                                ) -> None:
        """Update event filter parameters.

        This method is thread-safe.
        It will be called in the thread that owns the `PicoQuant_MultiHarp150`, `PicoQuant_HydraHarp400` instance.

        Parameters:
            channel_filter: A dictionary of channel number with their respective event filter modes.
            sync_aperture: [min, max] aperture times respective to the synchronization moment per channel.
        """
        with self._condition:
            self._event_filter_channels = channel_filter
            self._event_filter_aperture = sync_aperture
            self._event_filter.set_event_filter_config(channel_filter, sync_aperture)

    def set_histogram_config(self, channels: List[int], bin_resolution_ps: int, num_bins: int, num_sync: int) -> None:
        """Configure real-time histograms.

        This method is thread-safe.
        It will be called in the thread that owns the `PicoQuant_MultiHarp150`, `PicoQuant_HydraHarp400` instance.

        Parameters:
            channels: Channel numbers to collect the histogram for.
            bin_resolution_ps: The bin resolution in picoseconds.
            num_bins: The number of bins in the histogram.
            num_sync: The number of synchronizations to collect the histogram data for.
        """
        with self._condition:
            self._histogram_channels = channels
            self._histogram_resolution = bin_resolution_ps
            self._histogram_num_bins = num_bins
            self._histogram_num_sync = num_sync
            self._histogram_processor.set_histogram_config(channels, bin_resolution_ps, num_bins, num_sync)

    def set_countrate_config(self, sync_aperture: Tuple[int, int], num_sync: int) -> None:
        """Configure real-time count rate reporting.

        This method is thread-safe.
        It will be called in the thread that owns the `PicoQuant_MultiHarp150`, `PicoQuant_HydraHarp400` instance.

        Parameters:
            sync_aperture: [min, max] aperture times respective to the synchronization moment per channel.
            num_sync: The number of synchronizations to collect the count rate for.
        """
        with self._condition:
            self._countrate_aperture = sync_aperture
            self._countrate_num_sync = num_sync
            self._histogram_processor.set_countrate_config(sync_aperture, num_sync)

    def _request_shutdown(self) -> None:
        with self._condition:
            self._condition.notify_all()

    def get_events(self, max_events: int) -> Tuple[float, np.ndarray]:
        """Return events fetched by the background thread.

        Parameters:
            max_events: Maximum number of events to return.

        Returns:
            Tuple (timestamp, event_data).

        This method is thread-safe.
        It will be called in the thread that owns the `PicoQuant_MultiHarp150`, `PicoQuant_HydraHarp400` instance.
        """

        # Accept at most `max_events` pending event records.
        # This must be done under the condition variable lock.
        with self._condition:

            data_timestamp = self._data_timestamp

            if self._pending_events_overflow:
                # Events got dropped because there were too many pending events.
                # Discard all pending events and clear the overflow flag, then raise an exception.
                # TODO: Handle the events up to 'max' anyhow first and only then discard the rest?
                self._data = []
                self._num_pending_events = 0
                self._pending_events_overflow = False
                raise QMI_RuntimeException("Too many pending events from time tagger, pending events discarded.")

            if self._num_pending_events <= max_events:
                # Accept all pending events.
                data = self._data
                self._data = []
                self._num_pending_events = 0

            else:
                # Split the set of pending events.
                room_left = max_events
                split_pos = 0
                while True:
                    num_events = len(self._data[split_pos])
                    if num_events > room_left:
                        break
                    room_left -= num_events
                    split_pos += 1
                data = self._data[:split_pos]
                self._data = self._data[split_pos:]
                if room_left > 0:
                    data.append(self._data[0][:room_left])
                    self._data[0] = self._data[0][room_left:]
                self._num_pending_events -= max_events

        # Concatenate the event arrays to be returned.
        if len(data) == 0:
            event_data = np.empty(0, dtype=EventDataType)
        else:
            event_data = np.concatenate(data)

        return data_timestamp, event_data


class _EventFilter:
    """Filter T2 event records."""

    def __init__(self) -> None:
        """Initialize the event filter."""
        self._previous_sync_timestamp = -1
        self._previous_sync_reported = False
        self._sync_aperture = (0, 0)
        self._channel_filter_map = np.zeros(128, dtype=np.uint8)
        self._aperture_filter_enabled = False

    def set_event_filter_config(self,
                                channel_filter: Dict[int, EventFilterMode],
                                sync_aperture: Tuple[int, int]
                                ) -> None:
        """Set event filter parameters.

        Parameters:
            channel_filter: Mapping from event type (input channel) to the event filter mode for that event type.
            sync_aperture:  Tuple (delta_min, delta_max) in multiples of the instrument base resolution,
                            defining a time window following each SYNC event.
                            For event types that use `EventFilterMode.APERTURE`, only events that fall inside
                            this time window will pass the filter.
        """
        self._sync_aperture = sync_aperture
        self._channel_filter_map[:] = 0
        self._aperture_filter_enabled = False

        # Construct a map from event type code to channel filter mode.
        for (event_type, filter_mode) in channel_filter.items():
            self._channel_filter_map[event_type] = int(filter_mode)
            if filter_mode == EventFilterMode.APERTURE:
                self._aperture_filter_enabled = True

    def process_events(self, event_records: np.ndarray) -> np.ndarray:
        """Filter MultiHarp event records.

        This function filters a sequence of event records and returns only
        the events that are considered to be interesting according to the
        configured filter rules.

        The `_EventFilter` instance keeps track of the most recent SYNC event.
        The sequence of calls to this method must therefore correspond to
        a single TTTR event stream, processed in order, without missing events.

        Parameters:
            event_records: Numpy array of EventDataType records.

        Returns:
            Numpy array of event records that pass the filter.
        """

        # Do not waste time processing an empty array.
        if len(event_records) == 0:
            return event_records

        # Prepend the previous SYNC event, if there is any.
        # This simplifies processing of the aperture filter.
        previous_sync_inserted = False
        if self._aperture_filter_enabled and (self._previous_sync_timestamp >= 0):
            event_records = np.insert(event_records, 0, (SYNC_TYPE, self._previous_sync_timestamp))
            previous_sync_inserted = True

        # Map event type of each record to the corresponding filter mode.
        record_filter_modes = self._channel_filter_map[event_records["type"]]

        # Select events on channels that are unconditionally enabled.
        events_selected_bool = (record_filter_modes == EventFilterMode.ALL_EVENTS)

        # Find the indexes of SYNC events.
        (sync_events_idx,) = np.where(event_records["type"] == SYNC_TYPE)

        # Handle aperture filter, if necessary.
        if self._aperture_filter_enabled and (len(sync_events_idx) > 0):

            # For each event, determine time delta since the last SYNC event.
            # The result is only valid for event past the first SYNC.
            first_sync_idx = sync_events_idx[0]
            sync_deltas = _get_sync_deltas(event_records, sync_events_idx)

            # Mark all events on channels that are configured to use aperture filtering.
            events_aperture_bool = (record_filter_modes == EventFilterMode.APERTURE)

            # Reject events that do not pass the SYNC aperture.
            (delta_min, delta_max) = self._sync_aperture
            events_aperture_bool[:first_sync_idx] = False
            events_aperture_bool &= (sync_deltas >= delta_min)
            events_aperture_bool &= (sync_deltas <= delta_max)

            # Select events that pass the aperture filter.
            events_selected_bool |= events_aperture_bool

            # If aperture filtering is enabled on the SYNC channel, temporarily select all SYNC events.
            if self._channel_filter_map[SYNC_TYPE] == EventFilterMode.APERTURE:
                events_selected_bool[sync_events_idx] = True

        # If we prepended a previous, already reported SYNC event,
        # drop that event now to avoid reporting it a second time.
        if previous_sync_inserted and self._previous_sync_reported:
            events_selected_bool[0] = False

        # Update the timestamp of the last SYNC event.
        if len(sync_events_idx) > 0:
            self._previous_sync_timestamp = event_records[sync_events_idx[-1]]["timestamp"]
            self._previous_sync_reported = True

        # Discard rejected events.
        event_records = event_records[events_selected_bool]

        # If aperture filtering is enabled on the SYNC channel,
        # reject SYNC events unless they are followed by a non-SYNC event.
        if self._channel_filter_map[SYNC_TYPE] == EventFilterMode.APERTURE:

            # Select non-SYNC events.
            events_selected_bool = (event_records["type"] != SYNC_TYPE)

            # Also select all events that are followed by a non-SYNC event.
            events_selected_bool[np.where(events_selected_bool[1:])] = True

            # Remember whether the last SYNC event is dropped, so it can
            # be reported later if it becomes relevant again.
            if (len(event_records) > 0) and (event_records[-1]["type"] == SYNC_TYPE):
                self._previous_sync_reported = False

            # Discard rejected events.
            event_records = event_records[events_selected_bool]

        # Return the events that pass the filter.
        return event_records



"""
# QMI Process Manager

This tool starts or stops background QMI processes.
"""
import builtins
import sys
import argparse
import logging
import os
import os.path
import socket
import subprocess
from subprocess import Popen
import time

from typing import Callable, List, NamedTuple, Optional, Tuple

import colorama
import psutil

import qmi
from qmi.core.config_defs import CfgQmi
from qmi.core.exceptions import (
    QMI_Exception, QMI_ApplicationException, QMI_UnknownNameException,
    QMI_RpcTimeoutException
)
from qmi.core.util import format_address_and_port, is_valid_object_name


# Am I running in a windows virtual environment?
# sys.prefix and sys.base_prefix are not equal within a virtual environment.
# This is the documented way of validation: https://docs.python.org/3/library/venv.html#how-venvs-work.
WINENV = (sys.prefix != sys.base_prefix and os.name == 'nt')  # type: ignore

# Time in seconds to wait until a context shuts down after a request.
CONTEXT_SHUTDOWN_TIMEOUT = 8

# Default command to run on remote computer to start a process management server.
DEFAULT_SERVER_COMMAND = "python3 -m qmi.tools.proc server"

# Global variable holding the logger for this module.
if __name__ == "__main__":
    _logger = logging.getLogger("qmi_proc")

else:
    _logger = logging.getLogger(__name__)


# Result from shutdown_context().
ShutdownResult = NamedTuple("ShutdownResult", [
    ("responding", bool),   # True if the context was initially responding.
    ("pid", int),           # Process ID (if context responded).
    ("success", bool),      # True if shutdown was successful.
])


class ProcessException(Exception):
    """Raised when a process management operation fails."""
    pass


class ProcessManagementClient:
    """Client side of a remote process management server."""

    def __init__(self, host: str, context_name: str) -> None:
        """Connect to a remote computer via SSH and start a remote process management server.

        Parameters:
            host:           Host to connect to.
            context_name:   Name of context to manage.

        Raises:
            ProcessException: If connecting to the remote computer fails.
        """
        # Context name to manage with this instance.
        self._context_name = context_name

        # Get QMI configuration.
        cfg = qmi.context().get_config()

        # Get SSH host, user, command from configuration.
        server_command = None
        ssh_host = None
        ssh_user = None
        hostcfg = cfg.process_management.hosts.get(host)
        if hostcfg is not None:
            server_command = hostcfg.server_command
            ssh_host = hostcfg.ssh_host
            ssh_user = hostcfg.ssh_user

        # Fill in defaults
        if not server_command:
            server_command = DEFAULT_SERVER_COMMAND
        if not ssh_host:
            ssh_host = host

        # Check if we need to activate a virtual environment on the remote.
        assert context_name in cfg.contexts.keys()
        ctxcfg = cfg.contexts[context_name]
        venv_path = ctxcfg.virtualenv_path
        if venv_path is not None:
            server_command = os.path.join(venv_path, "bin", "") + server_command

        # Prepare command to invoke SSH.
        cmdline = [
            "ssh",                  # run SSH
            "-a",                   # disable agent forwarding
            "-x",                   # disable X11 forwarding
            "-T",                   # disable TTY allocation
            "-q",                   # suppress warnings
            "-o", "BatchMode yes"   # suppress password prompts
        ]
        if ssh_user:
            cmdline += ["-l", ssh_user]
        cmdline += [ssh_host, server_command]

        # Invoke SSH.
        _logger.debug("Running %s", cmdline)
        try:
            self._proc: Popen = subprocess.Popen(
                cmdline,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE)
        except (OSError, subprocess.SubprocessError) as exc:
            _logger.debug("Can not start SSH (%s: %s)", type(exc).__name__, str(exc))
            raise ProcessException("Can not start SSH ({}: {})"
                                   .format(type(exc).__name__, str(exc)))
        assert self._proc.stdin is not None
        assert self._proc.stdout is not None

    def close(self) -> None:
        """Close the remote server process."""

        assert self._proc.stdin is not None
        assert self._proc.stdout is not None

        # Close input stream to remote process.
        self._proc.stdin.close()

        # Wait until SSH process ends.
        try:
            self._proc.wait(timeout=2)
        except TimeoutError:
            _logger.debug("SSH client keeps running after stdin closed")
            self._proc.kill()

        self._proc.stdout.close()

    def start_process(self) -> int:
        """Request that the remote process management server start a new process."""

        assert self._proc.stdin is not None
        assert self._proc.stdout is not None

        # Send command to remote process management server.
        cmd = "START {}".format(self._context_name)
        self._proc.stdin.write(cmd.encode("ascii") + b"\n")
        self._proc.stdin.flush()

        # Receive response from server.
        resp = self._proc.stdout.readline()
        decoded_resp = resp.decode("ascii", errors="replace")

        # Handle response.
        if not decoded_resp:
            # Process server died.
            raise ProcessException("Remote process management failed")
        if decoded_resp.startswith("OK "):
            # Got result from process server.
            try:
                pid = int(decoded_resp[3:].strip())
            except ValueError:
                raise ProcessException("Remote process management protocol error")
            return pid
        elif decoded_resp.startswith("ERR "):
            # Got error from process server.
            msg = decoded_resp[4:].strip()
            raise ProcessException(msg)
        else:
            raise ProcessException("Invalid response from remote process manager ({!r})".format(decoded_resp))

    def stop_process(self, pid: int) -> bool:
        """Request that the remote process management server stop a running process."""

        assert self._proc.stdin is not None
        assert self._proc.stdout is not None

        # Send command to remote process management server.
        cmd = "STOP {} {}".format(self._context_name, pid)
        self._proc.stdin.write(cmd.encode("ascii") + b"\n")
        self._proc.stdin.flush()

        # Receive response from server.
        resp = self._proc.stdout.readline()
        decoded_resp = resp.decode("ascii", errors="replace")

        # Handle response.
        if not decoded_resp:
            # Process server died.
            raise ProcessException("Remote process management failed")
        if decoded_resp.startswith("OK "):
            # Got result from process server.
            try:
                success = int(decoded_resp[3:].strip())
            except ValueError:
                raise ProcessException("Remote process management protocol error")
            return success != 0
        elif decoded_resp.startswith("ERR "):
            # Got error from process server.
            msg = decoded_resp[4:].strip()
            raise ProcessException(msg)
        else:
            raise ProcessException("Invalid response from remote process manager ({!r})".format(decoded_resp))


def is_local_host(host: str) -> bool:
    """Return True if the host specification refers to the local computer."""

    # Resolve host.
    try:
        addrs = socket.getaddrinfo(host, None, type=socket.SOCK_STREAM)  # type: ignore
    except OSError:
        _logger.debug("Invalid host name '%s'", host)
        return False

    # Get list of IP addresses for the specified host.
    host_ips: set[str] = set()
    for addr in addrs:
        host_ips.add(str(addr[4][0]))

    # Return True if the specified host refers to a loopback address.
    for ip_address in host_ips:
        assert isinstance(ip_address, builtins.str), "'ip_address' unexpectedly not a string!"
        if ip_address.startswith("127.") or ip_address == "::1":
            return True

    # Get list of IP addresses matching the local computer.
    if_addrs = psutil.net_if_addrs()

    # Return True if a local IP address matches the specified host.
    for addrs in if_addrs.values():
        for addr in addrs:
            assert isinstance(addr, psutil._common.snicaddr)
            if addr.address in host_ips:
                return True

    # Specified host does not refer to the local computer.
    return False


def start_local_process(context_name: str) -> int:
    """Start the specified process on the local computer.

    :param context_name: Context to be started.
    :return: Process ID of the newly started process.
    :raises ProcessException: If the process can not be started.
    """

    _logger.debug("Starting local process for context %s", context_name)

    # Get context info from QMI configuration.
    cfg = qmi.context().get_config()
    ctxcfg = cfg.contexts.get(context_name)

    # Check that the context exists in the configuration.
    if ctxcfg is None:
        raise ProcessException("Unknown context '{}'".format(context_name))

    # Get program module and arguments from configuration.
    program_module = ctxcfg.program_module
    program_args = ctxcfg.program_args
    if not program_module:
        raise ProcessException("No program module configured for context '{}'".format(context_name))

    # Extend PYTHONPATH if needed.
    environment = os.environ.copy()
    python_path = ctxcfg.python_path
    if python_path is not None:
        if not os.path.isdir(python_path):
            raise ProcessException("PYTHONPATH is not a valid path for context {}".format(context_name))
        environment["PYTHONPATH"] = python_path

    # Check if a virtual environment needs to be activated.
    venv_path = ctxcfg.virtualenv_path
    if venv_path is not None:
        executable = os.path.join(venv_path, "bin", "python3")
    else:
        executable = sys.executable

    # Check that a host is configured for this context.
    if not ctxcfg.host:
        raise ProcessException("No host configured for context '{}'".format(context_name))

    # Find the directory where output logs will be written.
    output_dir = cfg.process_management.output_dir
    if output_dir:
        output_dir = qmi.context().resolve_file_name(output_dir)
    else:
        output_dir = qmi.context().get_qmi_home_dir()

    # Create output log file.
    datetime = time.strftime("%Y%m%d_%H%M%S", time.gmtime())
    output_file_name = os.path.join(output_dir, context_name + "_" + datetime + ".out")
    try:
        output_file = open(output_file_name, "a")
    except OSError as exc:
        raise ProcessException("Can not create output log file '{}' ({})".format(
            output_file_name, str(exc)))

    # Start a new Python process to run the specified program.
    # Close standard input.
    # Redirect standard output and stderr to the output log file.
    cmdline = [executable, "-m", program_module] + program_args
    try:
        proc = subprocess.Popen(cmdline,
                                stdin=subprocess.DEVNULL,
                                stdout=output_file,
                                stderr=subprocess.STDOUT,
                                start_new_session=True,
                                env=environment)
    except (OSError, subprocess.SubprocessError) as exc:
        raise ProcessException("Can not start program ({}: {})"
                               .format(type(exc).__name__, str(exc)))

    # Close output log file in parent process.
    output_file.close()

    # Wait and check if managed process is running.
    time.sleep(1)
    ret = proc.poll()
    if ret is not None:
        raise ProcessException("Program started but already stopped (status={})".format(ret))

    # Workaround for a limitation in the Python subprocess module.
    # The subprocess module does not allow a Popen object to be destroyed
    # while the child process is still running. By setting the "returncode"
    # attribute, we can make the subprocess module believe that the child
    # process has completed while it is actually still running.
    # See https://bugs.python.org/issue26741 which introduces this limitation.
    proc.returncode = 0

    pid = proc.pid
    if WINENV:
        parent = psutil.Process(pid)
        # always the only child, due to the nature in which this process is spawned.
        pid = parent.children()[0].pid

    # Return PID of managed process.
    return pid


def stop_local_process(context_name: str, pid: int) -> bool:
    """Stop the specified process on the local computer.

    :param context_name: Context to be stopped.
    :param pid: Process ID.
    :return: True if the process was stopped; False if the process was not running.
    :raises ProcessException: If the process can not be stopped.
    """

    _logger.debug("Stopping local process for context %s (PID=%d)", context_name, pid)

    # Search for process with specified process ID.
    try:
        proc = psutil.Process(pid=pid)
    except psutil.NoSuchProcess:
        # Process not running.
        _logger.debug("No process found with PID=%d", pid)
        return False

    # Check that process is running.
    try:
        if not proc.is_running():
            _logger.debug("Process with PID=%d is not running", pid)
            return False
    except psutil.Error as exc:
        # May happen in case of permission errors on Windows.
        raise ProcessException("Can not check process status ({})".format(type(exc).__name__))

    # Kill process.
    try:
        proc.kill()
    except psutil.NoSuchProcess:
        # Program disappeared before we could kill it.
        _logger.debug("Process with PID=%d disappeared before kill", pid)
        return False
    except psutil.Error as exc:
        # May happen in case of permission error.
        raise ProcessException("Can not kill process ({})".format(type(exc).__name__))

    # Wait until process ends.
    try:
        proc.wait(timeout=2)
        if proc.is_running():
            raise ProcessException("Process killed but still running (pid={})".format(proc.pid))
    except psutil.Error as exc:
        raise ProcessException("Can not check process status ({})".format(type(exc).__name__))

    # Process successfully stopped.
    return True


def start_process(context_name: str) -> int:
    """Start the specified process on a local or remote computer.

    :param context_name: Context to be started.
    :return: Process ID of the newly started process.
    :raises ProcessException: If the process can not be started.
    """

    # Get context info from QMI configuration.
    cfg = qmi.context().get_config()
    ctxcfg = cfg.contexts.get(context_name)

    # Check that the context exists in the configuration.
    if ctxcfg is None:
        raise ProcessException("Unknown context '{}'".format(context_name))

    # Check that a host is configured for this context.
    if not ctxcfg.host:
        raise ProcessException("No host configured for context '{}'".format(context_name))

    if is_local_host(ctxcfg.host):
        # Apply local process management.
        return start_local_process(context_name)
    else:
        # Apply remote process management.
        client = ProcessManagementClient(ctxcfg.host, context_name)
        try:
            return client.start_process()
        finally:
            client.close()


def stop_process(context_name: str, pid: int) -> bool:
    """Stop the specified process on a local or remote computer.

    :param context_name: Context to be stopped.
    :param pid: Process ID.
    :return: True if the process was stopped; False if the process was not running.
    :raises ProcessException: If the process can not be stopped.
    """

    # Get context info from QMI configuration.
    cfg = qmi.context().get_config()
    ctxcfg = cfg.contexts.get(context_name)

    # Check that the context exists in the configuration.
    if ctxcfg is None:
        raise ProcessException("Unknown context '{}'".format(context_name))

    # Check that a host is configured for this context.
    if not ctxcfg.host:
        raise ProcessException("No host configured for context '{}'".format(context_name))

    if is_local_host(ctxcfg.host):
        # Apply local process management.
        return stop_local_process(context_name, pid)
    else:
        # Apply remote process management.
        client = ProcessManagementClient(ctxcfg.host, context_name)
        try:
            return client.stop_process(pid)
        finally:
            client.close()


def get_context_status(context_name: str) -> Tuple[int, str]:
    """Check whether context is responding via TCP.

    :param context_name: Name of the context to be tested.
    :return: A tuple containing the process ID of the Python program containing the context, or -1 if the context is
        not responding via TCP. And the QMI version number of the newly created context.
    :raises ProcessException: If an error occurs.
    """

    # Get context info from QMI configuration.
    cfg = qmi.context().get_config()
    ctxcfg = cfg.contexts[context_name]

    # Check that the context is configured to support TCP connections.
    if (not ctxcfg.host) or (not ctxcfg.tcp_server_port):
        # Context does not support TCP connections.
        raise ProcessException("Context {} does not support TCP connections".format(context_name))

    # Try to connect to the context.
    peer_addr = format_address_and_port((ctxcfg.host, ctxcfg.tcp_server_port))
    try:
        qmi.context().suppress_version_mismatch_warnings = True
        qmi.context().connect_to_peer(context_name, peer_addr)

    except OSError as exc:
        # Can not connect to context; mark it as not responding.
        _logger.debug("Can not connect to context %r (%s: %s)",
                      context_name, type(exc).__name__, str(exc))
        return -1, ""

    except QMI_Exception as exc:
        # Unexpected error while connecting to context (bad handshake, etc.).
        _logger.debug("Protocol error from context %r (%s: %s)", context_name, type(exc).__name__, str(exc))
        raise ProcessException("Protocol error from context {}".format(context_name))

    # Successfully connected to peer context.
    # Get a proxy for the remote ContextInfo object.
    proxy = qmi.context().make_peer_context_proxy(context_name)

    # Get peer process ID.
    try:
        pid = proxy.get_pid()
    except QMI_Exception as exc:
        # Unexpected error
        _logger.debug("Error in get_pid() call to context %r (%s: %s)",
                      context_name, type(exc).__name__, str(exc))
        raise ProcessException("Can not get PID of context {}".format(context_name))

    # Get peer QMI version
    ver = proxy.get_version()

    # Disconnect from the peer context.
    try:
        qmi.context().disconnect_from_peer(context_name)
    except QMI_UnknownNameException:
        pass  # apparently already disconnected

    # Mark context as responding.
    return pid, ver


def shutdown_context(context_name: str, progressfn: Callable[[str], None]) -> ShutdownResult:
    """Send a shutdown request to the specified context.

    Send a soft shutdown request and wait until the context goes away.
    If that does not work, send a hard shutdown request and wait until
    the context goes away. If that does not work, give up and report failure.

    :param context_name: Context to shut down.
    :param progressfn: Callback function to report progress message.
    :return: Tuple (responding, pid, success).
    :raises ProcessException: If an error occurs.
    """
    def wait_disappear() -> bool:
        """Wait for a context to disappear.

        :return: Boolean to tell if context disappeared (True) or not (False).
        """
        t = 0.0
        while t < CONTEXT_SHUTDOWN_TIMEOUT:
            time.sleep(0.5)
            t += 0.5
            progressfn("")
            if not qmi.context().has_peer_context(context_name):
                # Peer context disappeared.
                return True

        return False

    # Get context info from QMI configuration.
    cfg = qmi.context().get_config()
    ctxcfg = cfg.contexts[context_name]

    # Check that the context is configured to support TCP connections.
    if (not ctxcfg.host) or (not ctxcfg.tcp_server_port):
        raise ProcessException("Context {} does not support TCP connections".format(context_name))

    # Supress warning for shutdown
    qmi.context().suppress_version_mismatch_warnings = True

    # Try to connect to the context.
    peer_addr = format_address_and_port((ctxcfg.host, ctxcfg.tcp_server_port))
    try:
        qmi.context().connect_to_peer(context_name, peer_addr)
    except OSError as exc:
        # Can not connect to context; mark it as not responding.
        _logger.debug("Can not connect to context %r (%s: %s)",
                      context_name, type(exc).__name__, str(exc))
        return ShutdownResult(responding=False, pid=-1, success=False)
    except QMI_Exception as exc:
        # Unexpected error while connecting to context (bad handshake, etc.).
        _logger.debug("Protocol error from context %r (%s: %s)", context_name, type(exc).__name__, str(exc))
        raise ProcessException("Protocol error from context {}".format(context_name))

    try:
        # Successfully connected to peer context and received handshake.
        # Get a proxy for the remote ContextInfo object.
        proxy = qmi.context().make_peer_context_proxy(context_name)

        # Get peer process ID.
        try:
            pid = proxy.get_pid()
        except QMI_Exception as exc:
            # Unexpected error
            _logger.debug("Error in get_pid() call to context %r (%s: %s)",
                          context_name, type(exc).__name__, str(exc))
            raise ProcessException("Can not get PID of context {}".format(context_name))

        # Send soft shutdown request.
        progressfn("soft shutdown")
        future = proxy.rpc_nonblocking.shutdown_context(hard=False)

        # Wait for answer from RPC call, but recover when no answer is received.
        try:
            future.wait(timeout=2)
        except QMI_RpcTimeoutException:
            _logger.debug("RPC timeout in soft shutdown to context %r", context_name)
            progressfn("timeout")
        except QMI_Exception as exc:
            # Unexpected error
            _logger.debug("Error in soft shutdown to context %r (%s: %s)",
                          context_name, type(exc).__name__, str(exc))
            raise ProcessException("Error in soft shutdown to context {}".format(context_name))

        # Wait until context disappears.
        if wait_disappear():
            return ShutdownResult(responding=True, pid=pid, success=True)

        # Soft shutdown failed.
        # Send hard shutdown request.
        progressfn("hard shutdown")
        future = proxy.rpc_nonblocking.shutdown_context(hard=True)

        # We don't expect an answer from the RPC call.
        # Clean up the future.
        try:
            future.wait(timeout=0)
        except QMI_RpcTimeoutException:
            pass  # ignore timeout

        # Wait until context disappears.
        if wait_disappear():
            return ShutdownResult(responding=True, pid=pid, success=True)

        # Hard shutdown failed.
        return ShutdownResult(responding=True, pid=pid, success=False)

    finally:
        # Disconnect from peer context.
        try:
            qmi.context().disconnect_from_peer(context_name)
        except QMI_UnknownNameException:
            pass  # apparently already disconnected


def select_context_by_name(cfg: CfgQmi, context_name: str) -> List[str]:
    """Return input the context name in a list if it is valid.

    Parameters:
        cfg: The current QMI configuration.
        context_name: The name of the context to validate.

    Raises:
        QMI_ApplicationException: If context name is not within known contexts or name is invalid.

    Returns:
        The input context name in a list.
    """
    # Return only the specified context.
    if context_name not in cfg.contexts:
        raise QMI_ApplicationException("Unknown context name '{}'".format(context_name))

    if not is_valid_object_name(context_name):
        raise QMI_ApplicationException("Invalid context name '{}'".format(context_name))

    return [context_name]


def select_contexts(cfg: CfgQmi) -> List[str]:
    """Return a list of all applicable context names.

    Parameters:
        cfg: The current QMI configuration.

    Raises:
        QMI_ApplicationException: If no contexts are enabled in the configuration, or name is invalid.

    Returns:
        All context names as a list.
    """
    context_names = []

    # Return all enabled contexts.
    for (ctxname, ctxcfg) in cfg.contexts.items():
        if ctxcfg.enabled:
            context_names.append(ctxname)
    if not context_names:
        raise QMI_ApplicationException("There are no enabled contexts in the configuration")

    # Sanity check on context names.
    # Note that our remote process management protocol does not correctly
    # handle context names containing whitespace or non-ASCII characters.
    for ctxname in context_names:
        if not is_valid_object_name(ctxname):
            raise QMI_ApplicationException("Invalid context name '{}'".format(ctxname))

    return context_names


def select_local_contexts(cfg: CfgQmi) -> List[str]:
    """Return a list of all applicable local context names.

    Parameters:
        cfg: The current QMI configuration.

    Raises:
        QMI_ApplicationException: If no local contexts are enabled in the configuration, or name is invalid.

    Returns:
        Local context names as a list.
    """
    context_names = []

    # Return all enabled local contexts.
    for (ctxname, ctxcfg) in cfg.contexts.items():
        if ctxcfg.enabled and ctxcfg.host and is_local_host(ctxcfg.host):
            context_names.append(ctxname)

    if not context_names:
        raise QMI_ApplicationException("There are no enabled local contexts in the configuration")

    # Sanity check on context names.
    # Note that our remote process management protocol does not correctly
    # handle context names containing whitespace or non-ASCII characters.
    for ctxname in context_names:
        if not is_valid_object_name(ctxname):
            raise QMI_ApplicationException("Invalid context name '{}'".format(ctxname))

    return context_names


def show_progress_msg(msg: str) -> None:
    """Print short progress message on screen."""
    if msg:
        print("(" + msg + ")", end=" ")
    else:
        print(".", end=" ")
    sys.stdout.flush()


def proc_server(cfg: CfgQmi) -> int:
    """Run the process manager in server mode.

    Parameters:
        cfg: The current QMI configuration.

    Raises:
        ProcessException: If process is unknown or should not run on this host.
        ValueError: By invalid command.

    Returns:
        Exit status (0 = success).
    """

    _logger.debug("Running qmi_proc in server mode")

    ret = 0

    # Serve remote requests via stdin/stdout until EOF on stdin.
    while True:
        # Get next request via stdin.
        sys.stdout.flush()
        req = sys.stdin.readline()

        # Stop when EOF is seen on stdin.
        if not req:
            break

        try:
            # Parse request.
            words = req.strip().split()

            if len(words) == 2 and words[0] == "START":
                # Received command to start a new process.
                context_name = words[1]

                # Get context info from QMI configuration.
                ctxcfg = cfg.contexts.get(context_name)
                if ctxcfg is None:
                    raise ProcessException("Unknown context")

                # Check that context is configured to run on the local host.
                if (not ctxcfg.host) or (not is_local_host(ctxcfg.host)):
                    raise ProcessException("Context should not run on this host")

                # Start process.
                pid = start_local_process(context_name)

                # Report result.
                print("OK {}".format(pid))
                sys.stdout.flush()

            elif len(words) == 3 and words[0] == "STOP":
                # Received command to stop a running process.
                context_name = words[1]
                pid = int(words[2])

                # Sanity check on PID.
                if pid < 1:
                    raise ValueError("Invalid request")

                # Get context info from QMI configuration.
                ctxcfg = cfg.contexts[context_name]
                if ctxcfg is None:
                    raise ProcessException("Unknown context")

                # Check that context is configured to run on the local host.
                if (not ctxcfg.host) or (not is_local_host(ctxcfg.host)):
                    raise ProcessException("Context should not run on this host")

                # Stop process.
                result = stop_local_process(context_name, pid)

                # Report result.
                print("OK {}".format(int(result)))
                sys.stdout.flush()

            else:
                raise ValueError("Invalid request")

        except ProcessException as exc:
            # Error occurred while processing command.
            # Report error and continue to next command.
            print("ERR {}".format(exc))
            sys.stdout.flush()

        except ValueError:
            # Invalid command received from client.
            # Report error and abort server.
            _logger.debug("Invalid request %r", req)
            print("ERR Invalid request")
            sys.stdout.flush()
            ret = 1
            break

    return ret


def proc_start(cfg: CfgQmi, context_name: Optional[str], local: bool) -> int:
    """Start one or more processes.

    Parameters:
        cfg:          The current QMI configuration.
        context_name: Process to start, or None to start all configured processes.
        local:        Boolean flag to indicate if we only look at local processes (True) or all processes (False).

    Raises:
        ProcessException: By unexpected PID number in check.

    Returns:
        Exit status (0 = success).
    """
    colorama.init()
    started_str = "[" + colorama.Fore.GREEN + "STARTED" + colorama.Fore.RESET + "]"
    failed_str = "[" + colorama.Fore.RED + "FAILED" + colorama.Fore.RESET + "]"

    ret = 0

    # Select applicable contexts.
    if context_name:
        context_names = select_context_by_name(cfg, context_name)
    elif local:
        context_names = select_local_contexts(cfg)
    else:
        context_names = select_contexts(cfg)

    print("Starting QMI processes:")

    # Process each applicable context.
    for context_name in context_names:
        # Show process name.
        print("    {:30s}:".format(context_name), end=" ")
        sys.stdout.flush()

        try:
            # Perhaps the process is already running.
            # Check if the peer context responds via TCP.
            pid, ver = get_context_status(context_name)
            if pid >= 0:
                print("already running (PID={}, QMI={})".format(pid, ver))

            else:
                # Context not responding via TCP.
                # Note that the program may still be running but just not responding.
                # We can not really do anything about that, so just try to start it and hope for the best.
                show_progress_msg("starting")
                pid = start_process(context_name)
                show_progress_msg("started PID={}".format(pid))

                # Check that newly started process responds via TCP.
                time.sleep(0.5)
                status_pid, _ = get_context_status(context_name)
                if status_pid < 0:
                    show_progress_msg("")
                    time.sleep(1.5)
                    status_pid, _ = get_context_status(context_name)

                if status_pid == pid:
                    # New process responds via TCP; everything OK.
                    show_progress_msg("responding")
                    print(started_str)
                elif status_pid < 0:
                    # New process does not respond via TCP.
                    print("not responding via TCP", failed_str)
                    ret = 1
                else:
                    raise ProcessException("New process (PID={}) for context {} reports unexpected PID={}"
                                           .format(pid, context_name, status_pid))

        except ProcessException as exc:
            print(failed_str)
            print("ERROR: {}".format(exc), file=sys.stderr)
            ret = 1

    print()
    colorama.deinit()

    return ret


def proc_stop(cfg: CfgQmi, context_name: Optional[str], local: bool) -> int:
    """Stop one or more running processes.

    Parameters:
        cfg:          The current QMI configuration.
        context_name: Process to start, or None to stop all configured processes.
        local:        Boolean flag to indicate if we only look at local processes (True) or all processes (False).

    Raises:
        ProcessException: By unexpected error in stopping the process.

    Returns:
        Exit status (0 = success).
    """
    colorama.init()
    stopped_str = "[" + colorama.Fore.GREEN + "STOPPED" + colorama.Fore.RESET + "]"
    failed_str = "[" + colorama.Fore.RED + "FAILED" + colorama.Fore.RESET + "]"

    ret = 0

    # Select applicable contexts.
    if context_name:
        context_names = select_context_by_name(cfg, context_name)
    elif local:
        context_names = select_local_contexts(cfg)
    else:
        context_names = select_contexts(cfg)

    print("Stopping QMI processes:")

    # Process each applicable context. Stop should happen in inverse order to start, in case of dependencies.
    for context_name in reversed(context_names):
        # Show process name.
        print("    {:30s}:".format(context_name), end=" ")
        sys.stdout.flush()

        try:
            # Try to shut down context via TCP.
            result = shutdown_context(context_name, show_progress_msg)
            if result.success:
                print(stopped_str)
            elif not result.responding:
                print("not responding via TCP")
            else:
                print(failed_str)
                # Failed to stop via TCP.
                # Try to stop process via local process management.
                print("    {:30s} ".format(""), end=" ")
                show_progress_msg("kill")
                sys.stdout.flush()
                if stop_process(context_name, result.pid):
                    print(stopped_str)
                else:
                    print("not running")

        except ProcessException as exc:
            print(failed_str)
            print("ERROR: {}".format(exc), file=sys.stderr)
            ret = 1

    print()
    colorama.deinit()

    return ret


def proc_status(cfg: CfgQmi, context_name: Optional[str]) -> int:
    """Show the status of one or more processes.

    Parameters:
        cfg:          The current QMI configuration.
        context_name: Process to get status for, or None to get all configured processes.

    Raises:
        ProcessException: By unexpected error when checking the status.

    Returns:
        Exit status (0 = success).
    """

    colorama.init()
    running_str = "[" + colorama.Fore.GREEN + "RUNNING" + colorama.Fore.RESET + "]"
    offline_str = "[" + colorama.Fore.RED + "OFFLINE" + colorama.Fore.RESET + "]"

    ret = 0

    # Select applicable contexts.
    if context_name:
        context_names = select_context_by_name(cfg, context_name)
    else:
        context_names = select_contexts(cfg)

    max_len_context_name = len(max(context_names, key=len))

    print("QMI process status:")

    # Process each applicable context.
    for context_name in context_names:
        print("    {:{max}s}:".format(context_name, max=max_len_context_name), end=" ")
        sys.stdout.flush()

        # Check if the peer context responds via TCP.
        try:
            pid, ver = get_context_status(context_name)
            # Print process status
            if pid >= 0:
                print(running_str, "responding via TCP (PID={}, QMI={})".format(pid, ver))
            else:
                print(offline_str, "not responding via TCP")
        except ProcessException as exc:
            print()
            print("ERROR: {}".format(exc), file=sys.stderr)
            ret = 1

    print()
    colorama.deinit()

    return ret


def main() -> int:
    """Main routine of QMI process manager.

    :return: 0 in case of success; exit status in case of error.
    """
    parser = argparse.ArgumentParser()
    parser.description = """This tool starts or stops background QMI processes.
        Processes are identified by their context name, as specified in
        the QMI configuration file. Processes can run either on the local computer
        or on a remote, network-connected computer."""
    parser.add_argument("--config", action="store", type=str,
                        help="specify the QMI configuration file")
    mutex_group = parser.add_mutually_exclusive_group()
    mutex_group.add_argument("--all", action="store_true", help="start or stop all configured processes")
    mutex_group.add_argument("--locals", action="store_true", help="start or stop local configured processes")
    parser.add_argument("command", action="store", choices=["start", "stop", "restart", "status", "server"],
                        help="'start' to start the specified process; 'stop' to stop a running process;"
                             + " 'restart' to restart the running process; 'status' to show the process status")
    parser.add_argument("context_name", action="store", type=str, nargs="?",
                        help="context name of the process to start or stop")
    args = parser.parse_args()

    if args.command == "server" and (args.all or args.context_name or args.locals):
        print("ERROR: Can not specify server mode together with other options")
        return 1

    if args.all and args.context_name:
        print("ERROR: Can not specify a context_name together with --all", file=sys.stderr)
        return 1

    if args.locals and args.context_name:
        print("ERROR: Can not specify a context_name together with --locals", file=sys.stderr)
        return 1

    if (args.command in ("start", "stop", "restart")) and (not args.all) and (not args.locals) and \
            (not args.context_name):
        print("ERROR: Specify either: a context_name, or --all, or --local", file=sys.stderr)
        return 1

    qmi.start("proc_mgr", config_file=args.config, console_loglevel="WARNING")
    # Get the QMI configuration.
    cfg = qmi.context().get_config()
    try:

        if args.command == "server":
            return proc_server(cfg=cfg)

        if args.command == "start":
            return proc_start(cfg=cfg, context_name=args.context_name, local=args.locals)

        elif args.command == "stop":
            return proc_stop(cfg=cfg, context_name=args.context_name, local=args.locals)

        elif args.command == "restart":
            proc_stop(cfg=cfg, context_name=args.context_name, local=args.locals)
            return proc_start(cfg=cfg, context_name=args.context_name, local=args.locals)

        elif args.command == "status":
            return proc_status(cfg=cfg, context_name=args.context_name)

        else:
            print("ERROR: Unknown command {!r}".format(args.command), file=sys.stderr)
            return 1

    except QMI_ApplicationException as exc:
        print("ERROR:", exc, file=sys.stderr)
        return 1

    finally:
        qmi.stop()


if __name__ == "__main__":
    sys.exit(main())



"""
InstruTech Inc., vacuum gauge controller.

The qmi.instruments.instru_tech package provides support for:
- AGC302 active vacuum gauge controller.
"""
# Alternative, QMI naming convention approved name
from qmi.instruments.instru_tech.instrutech_agc302 import InstruTech_AGC302 as InstruTech_Agc302



"""Unit test for AimTTi TGF3000 and TGF4000 series instrument driver."""

import math
import time
from typing import cast
import unittest
from unittest.mock import MagicMock, call, patch

import qmi
from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.transport import QMI_TcpTransport
from qmi.instruments.tt import AimTTi_Tgf30004000, WaveformType, CounterInputChannel


class TestTGF(unittest.TestCase):

    def setUp(self):
        qmi.start("Test_tgf_3000_4000")
        self._transport_mock = MagicMock(spec=QMI_TcpTransport)
        with patch(
                'qmi.instruments.tt.tgf.create_transport',
                return_value=self._transport_mock):
            self.instr: AimTTi_Tgf30004000 = qmi.make_instrument("instr", AimTTi_Tgf30004000, "transport_descriptor")
            self.instr = cast(AimTTi_Tgf30004000, self.instr)

    def tearDown(self):
        qmi.stop()

    def _open_helper(self):
        """Open instrument and check transport calls."""
        self.instr.open()
        self._transport_mock.open.assert_called_once_with()
        self._transport_mock.write.assert_called_once_with(b"*CLS\n")
        self._transport_mock.open.reset_mock()
        self._transport_mock.write.reset_mock()

    def _prep_cmd_no_error(self):
        """Prepare transport mock for error check without error."""
        self._transport_mock.read_until.return_value = b"+0\r\n"
        self._transport_mock.write.reset_mock()

    def _assert_cmds(self, expect_commands):
        """Assert that the driver sent the specified commands."""
        expect_calls = [call(cmd) for cmd in expect_commands]
        expect_calls.append(call(b"EER?\n"))
        self.assertEqual(self._transport_mock.write.mock_calls, expect_calls)

    def test_open_close(self):
        self._open_helper()
        self.instr.close()
        self._transport_mock.close.assert_called_once_with()

    def test_open_comm_error(self):
        self._transport_mock.write.side_effect = OSError("fake error")
        with self.assertRaises(OSError):
            self.instr.open()
        self._transport_mock.open.assert_called_once_with()
        self._transport_mock.write.assert_called_once_with(b"*CLS\n")
        self._transport_mock.close.assert_called_once_with()

    def test_get_idn(self):
        self._open_helper()
        self._transport_mock.read_until.return_value = b"vendor,model,serial,version\r\n"
        idn = self.instr.get_idn()
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"*IDN?\n")
        self.assertEqual(idn.vendor, "vendor")
        self.assertEqual(idn.model, "model")
        self.assertEqual(idn.serial, "serial")
        self.assertEqual(idn.version, "version")

    def test_get_idn_fail(self):
        self._open_helper()
        self._transport_mock.read_until.return_value = b"nonsense\r\n"
        with self.assertRaises(QMI_InstrumentException):
            idn = self.instr.get_idn()
        self.instr.close()

    def test_reset(self):
        self._open_helper()
        self._prep_cmd_no_error()
        self.instr.reset()
        self.instr.close()
        self._assert_cmds([b"*CLS\n", b"*RST\n"])

    def test_set_output_enabled(self):
        self._open_helper()

        self._prep_cmd_no_error()
        self.instr.set_output_enabled(1, True)
        self._assert_cmds([b"CHN 1;OUTPUT ON\n"])

        self._prep_cmd_no_error()
        self.instr.set_output_enabled(2, False)
        self.instr.close()
        self._assert_cmds([b"CHN 2;OUTPUT OFF\n"])

    def test_set_output_enabled_fail(self):
        self._open_helper()
        self._transport_mock.read_until.return_value = b"-111\r\n"
        with self.assertRaises(QMI_InstrumentException):
            self.instr.set_output_enabled(1, True)
        self.instr.close()
        self.assertEqual(self._transport_mock.write.mock_calls, [
            call(b"CHN 1;OUTPUT ON\n"),
            call(b"EER?\n")
        ])

    def test_nonsense_error(self):
        self._open_helper()
        self._transport_mock.read_until.return_value = b"nonsense\r\n"
        with self.assertRaises(QMI_InstrumentException):
            self.instr.set_output_enabled(1, True)
        self.instr.close()
        self.assertEqual(self._transport_mock.write.mock_calls, [
            call(b"CHN 1;OUTPUT ON\n"),
            call(b"EER?\n")
        ])

    def test_set_output_inverted(self):
        self._open_helper()

        self._prep_cmd_no_error()
        self.instr.set_output_inverted(1, True)
        self._assert_cmds([b"CHN 1;OUTPUT INVERT\n"])

        self._prep_cmd_no_error()
        self.instr.set_output_inverted(2, False)
        self.instr.close()
        self._assert_cmds([b"CHN 2;OUTPUT NORMAL\n"])

    def test_set_output_load_impedance(self):
        self._open_helper()

        self._prep_cmd_no_error()
        self.instr.set_output_load_impedance(1, 50.0)
        self._assert_cmds([b"CHN 1;ZLOAD 50\n"])

        self._prep_cmd_no_error()
        self.instr.set_output_load_impedance(2, math.inf)
        self.instr.close()
        self._assert_cmds([b"CHN 2;ZLOAD OPEN\n"])

    def test_set_waveform(self):
        self._open_helper()

        self._prep_cmd_no_error()
        self.instr.set_waveform(1, WaveformType.SINE)
        self._assert_cmds([b"CHN 1;WAVE SINE\n"])

        self._prep_cmd_no_error()
        self.instr.set_waveform(2, WaveformType.SQUARE)
        self.instr.close()
        self._assert_cmds([b"CHN 2;WAVE SQUARE\n"])

    def test_set_frequency(self):
        self._open_helper()
        self._prep_cmd_no_error()
        self.instr.set_frequency(1, 12345678.123456)
        self.instr.close()
        self._assert_cmds([b"CHN 1;FREQ 12345678.123456\n"])

    def test_set_square_duty_cycle(self):
        self._open_helper()
        self._prep_cmd_no_error()
        self.instr.set_square_duty_cycle(1, 12.345)
        self.instr.close()
        self._assert_cmds([b"CHN 1;SQRSYMM 12.345\n"])

    def test_set_ramp_symmetry(self):
        self._open_helper()
        self._prep_cmd_no_error()
        self.instr.set_ramp_symmetry(1, 34.567)
        self.instr.close()
        self._assert_cmds([b"CHN 1;RMPSYMM 34.567\n"])

    def test_set_amplitude(self):
        self._open_helper()
        self._prep_cmd_no_error()
        self.instr.set_amplitude(1, 2.345, -0.678)
        self.instr.close()
        self._assert_cmds([b"CHN 1;AMPL 2.345;DCOFFS -0.678\n"])

    def test_set_phase(self):
        self._open_helper()
        self._prep_cmd_no_error()
        self.instr.set_phase(1, 123.456)
        self.instr.close()
        self._assert_cmds([b"CHN 1;PHASE 123.456\n"])

    def test_get_external_reference_enabled(self):
        self._open_helper()

        self._transport_mock.read_until.return_value = b"INT\r\n"
        ret = self.instr.get_external_reference_enabled()
        self.assertFalse(ret)
        self._transport_mock.write.assert_called_once_with(b"CLKSRC?\n")

        self._transport_mock.read_until.return_value = b"EXT\r\n"
        ret = self.instr.get_external_reference_enabled()
        self.assertTrue(ret)

        self._transport_mock.read_until.return_value = b"ehm\r\n"
        with self.assertRaises(QMI_InstrumentException):
            ret = self.instr.get_external_reference_enabled()
        self.instr.close()

    def test_set_external_reference_enabled(self):
        self._open_helper()
        self._prep_cmd_no_error()
        self.instr.set_external_reference_enabled(True)
        self.instr.close()
        self._assert_cmds([b"CLKSRC EXT\n"])

    def test_set_counter_enabled(self):
        self._open_helper()
        self._prep_cmd_no_error()
        self.instr.set_counter_enabled(True)
        self.instr.close()
        self._assert_cmds([b"CNTRSWT ON\n"])

    def test_set_counter_input(self):
        self._open_helper()
        self._prep_cmd_no_error()
        self.instr.set_counter_input(CounterInputChannel.TRIG_IN)
        self.instr.close()
        self._assert_cmds([b"CNTRCPLNG DC\n"])

    def test_read_frequency(self):
        self._open_helper()
        self._transport_mock.read_until.side_effect = [b"125.001Hz\r\n", b"0.25\r\n", b"10x4Hz\r\n"]

        fake_time = time.time()
        with patch("qmi.instruments.tt.tgf.time.time", return_value=fake_time):
            ret = self.instr.read_frequency()
        self.assertEqual(ret.timestamp, fake_time)
        self.assertEqual(ret.frequency, 125.001)
        self._transport_mock.write.assert_called_once_with(b"CNTRVAL?\n")

        with self.assertRaises(QMI_InstrumentException):
            ret = self.instr.read_frequency()

        with self.assertRaises(QMI_InstrumentException):
            ret = self.instr.read_frequency()

        self.instr.close()

    def test_set_local(self):
        self._open_helper()
        self.instr.set_local()
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"LOCAL\n")

    def test_bad_channel(self):
        self._open_helper()
        with self.assertRaises(ValueError):
            self.instr.set_output_enabled(0, True)
        with self.assertRaises(ValueError):
            self.instr.set_frequency(3, 100.0e3)
        self.instr.close()


if __name__ == '__main__':
    unittest.main()
